/***** dataflow.h  --  Data flow analysis routines header file *****/

/* Data flow engine written by Radu Iosif <iosif@athena.polito.it>         */

/* dSpin extension written by Radu Iosif. Feedback: iosif@athena.polito.it */


#ifndef __DATAFLOW_H__
#define __DATAFLOW_H__


#include "spin.h"


/* Basic block definition */

typedef struct bblock {
  int            bid;          /* block id */
  int            mark;         /* dfs marker  */

  struct superset *in;         /* alias in set   */
  struct superset *out;        /* alias out set  */

  struct superset *in_ln;      /* 2nd degree alias in set  */
  struct superset *out_ln;     /* 2nd degree alias out set */

  int              e_no;       /* number of elements   */
  struct Element **e_seq;      /* sequence of elements */

  int             pred_no;     /* number of predecessors */
  struct bblock **pred_list;   /* list of predecessors   */

  int             succ_no;     /* number of successors   */
  struct bblock **succ_list;   /* list of successors     */
} * bb_t;


/* List of basic blocks */

typedef struct bb_list {
  int   b_no;
  bb_t *b_ls;
} * bl_t;

#define LS(x)    ((x)->b_ls)
#define NO(x)    ((x)->b_no)


/* Generic set definition */

#define WTYPE       unsigned char
#define WSIZE       sizeof(WTYPE)

typedef struct set {
  int    max;           /* max number of elements  */
  int    size;          /* size of bitmap in words */
  WTYPE *bmap;          /* bitmap                  */
} * set_t;


/* Superset definition */

typedef struct superset {
  int    s_no;                  /* number of sets   */
  set_t *s_set;                 /* list of sets     */
} *ss_t;


/* Variable descriptor */

typedef struct var_tag {
  struct set    *set;        /* name descriptor set */
  struct Lextok *tok;        /* the name lextok     */
  int            ptr;        /* is-a-pointer flag   */
  int            fptr;       /* is-a-fp flag        */
  int            glob;       /* is-global flag      */
} * var_t;


/* Function definition descriptor */

typedef struct fun_tag {
  struct Defun *fun;         /* function definition      */

  struct superset *in;       /* alias in set             */
  struct superset *in_ln;    /* 2nd degree alias in set  */
  struct superset *out;      /* alias out set            */
  struct set      *rset;     /* set of returned variables */
} * fun_t;


/* Link (2nd degree alias) descriptor */

typedef struct link {
  int l_no;                 /* number of links */
 
  struct ln_tag {
    int   idx;              /* variable index          */
    set_t set;              /* variable descriptor set */
  } *l_list;

} *ln_t;


#endif /* __DATAFLOW_H__ */
