/***** spin: pangen2.c *****/

/* Copyright (c) 1991-1998 by Lucent Technologies - Bell Laboratories     */
/* All Rights Reserved.  This software is for educational purposes only.  */
/* Permission is given to distribute this code provided that this intro-  */
/* ductory message is not removed and no monies are exchanged.            */
/* No guarantee is expressed or implied by the distribution of this code. */
/* Software written by Gerard J. Holzmann as part of the book:            */
/* `Design and Validation of Computer Protocols,' ISBN 0-13-539925-4,     */
/* Prentice Hall, Englewood Cliffs, NJ, 07632.                            */
/* Send bug-reports and/or questions to: gerard@research.bell-labs.com    */

/* dSpin extension written by Radu Iosif. Feedback: iosif@athena.polito.it */


#include "spin.h"
#include "version.h"
#ifdef PC
#include "y_tab.h"
#else
#include "y.tab.h"
#endif
#include "pangen2.h"


#define DELTA	500	/* sets an upperbound on nr of chan names */

#define blurb(fd, e)	fprintf(fd, "\n\t\t/* %s:%d */\n", \
				e->n->fn->name, e->n->ln)
#define tr_map(m, e)	fprintf(tt, "\t\ttr_2_src(%d, %s, %d);\n", \
				m, e->n->fn->name, e->n->ln);

extern ProcList	*rdy;
extern RunList	*run;
extern Defun    *dfs;

ProcList *Proc;

extern Symbol	*Fname, *oFname, *owner, *context, *RetType;
extern char	*claimproc, *eventmap;
extern int	lineno, verbose, Npars, Mpars, nrPars;
extern int	m_loss, Nid, has_remote;
extern int	Ntimeouts, Etimeouts;
extern int	u_sync, u_async, nrRdy, nrFun, nqs;
extern int	GenCode, IsGuard, Level, TestOnly, Expand_Ok;
extern char	*NextLab[];
extern SeqList  *cur_s;

extern int      doformal(FILE *, int, Sequence *, int, char *);
extern int      dolocal(FILE *, int, Sequence *, int);

FILE	*tc, *th, *tt, *tm, *tb;
int	OkBreak = -1;
short	nocast=0;	/* to turn off casts in lvalues */
short	terse=0;	/* terse printing of varnames */
short	no_arrays=0;
short	has_last=0;	/* spec refers to _last */
short	has_badelse=0;	/* spec contains else combined with chan refs */
short	has_enabled=0;	/* spec contains enabled() */
short	has_pcvalue=0;	/* spec contains pc_value() */
short	has_np=0;	/* spec contains np_ */
short	has_sorted=0;	/* spec contains `!!' (sorted-send) operator */
short	has_random=0;	/* spec contains `??' (random-recv) operator */
short	has_xu=0;	/* spec contains xr or xs assertions */
short	has_unless=0;	/* spec contains unless statements */
short	has_provided=0;	/* spec contains PROVIDED clauses on procs */
int	mst=0;		/* max nr of state/process */
int	claimnr = -1;	/* claim process, if any */
int	eventmapnr = -1; /* event trace, if any */
int	Pid;		/* proc currently processed */
int     heap_init_size; /* needed to declare iniglobals() */

static Lextok	*Nn[2];
static int	Det;	/* set if deterministic */
static int	T_sum, T_mus, t_cyc;
static int	TPE[2], EPT[2];
static int	uniq=1;
static short	withprocname=0;	/* prefix local varnames with procname */
static short	_isok=0;	/* checks usage of predefined varble _ */
static short	AllGlobal=0;	/* set if process has provided clause */

static int	has_global(Element *, Lextok *);
static int	getweight(Lextok *);
static int	scan_seq(Sequence *);
static void	genconditionals(void);
static void	mark_seq(Sequence *);
static void	patch_atomic(Sequence *);
static void	putproc(ProcList *);
static void	Tpe(Lextok *);
static void     putoffset(FILE *, Lextok *, Symbol *, int, long, int);

void    	put_seq(Sequence *, int, int);

extern void     putfun(Defun *);
extern void     genfunpars();
extern void     genpushpars();

extern void     dfa_all(ProcList *, Defun *);
extern int      global_alias(Element *, Lextok *);

extern int      *plen_tab;
extern int      **pmap_tab;

extern int      *clen_tab;
extern int      **cmap_tab;

extern int      typ_ck(int, int, char *);
extern void     alldone(int);

static int
fproc(char *s)
{	
  ProcList *p;
  
  for (p = rdy; p; p = p->nxt)
    if (strcmp(p->n->name, s) == 0)
      return p->tn;

  fatal("proctype %s not found", s);
  return -1;
}


static ProcList *
getproc(Symbol *s)
{
  ProcList *p;
  
  for (p = rdy; p; p = p->nxt)
    if (strcmp(p->n->name, s->name) == 0)
      return p;

  fatal("proctype %s not found", s->name);
  return (ProcList *)0;
}


static void reverse_procs(RunList *q)
{
  if (!q) 
    return;

  reverse_procs(q->nxt);
  fprintf(tc, "  Addproc(%d);\n", q->tn);
}


static void tm_predef_np(void)
{
  fprintf(th, "#define _T5	%d\n", uniq++);
  fprintf(th, "#define _T2	%d\n", uniq++);

  fprintf(tm, "\tcase  _T5:\t/* np_ */\n");
  fprintf(tm, "\t\tif (!((!(trpt->o_pm & 4) && !(trpt->tau & 128))))\n");
  fprintf(tm, "\t\t\tcontinue;\n");
  fprintf(tm, "\t\t/* else fall through */\n");
  fprintf(tm, "\tcase  _T2:\t/* true */\n");
  fprintf(tm, "\t\tm = 3; goto P999;\n");
}


static void tt_predef_np(void)
{
  fprintf(tt, "\t/* np_ demon: */\n");
  fprintf(tt, "\ttrans[_NP_] = ");
  fprintf(tt, "(Trans **) emalloc(2*sizeof(Trans *));\n");
  fprintf(tt, "\tT = trans[_NP_][0] = ");
  fprintf(tt, "settr(9997, 0, 0, _T2, 0, \"(1)\", 0, 2, 0);\n");
  fprintf(tt, "\tT->nxt = ");
  fprintf(tt, "settr(9998, 0, 1, _T5, 0, \"(np_)\", 1, 2, 0);\n");
  fprintf(tt, "\tT = trans[_NP_][1] = ");
  fprintf(tt, "settr(9999, 0, 1, _T5, 0, \"(np_)\", 1, 2, 0);\n");
}


void gensrc(void)
{
  ProcList *p;
  Defun    *d;
  int i, j, len_max = 0;
  
  if (!(tc = fopen("pan.c", "w"))	/* main routines */
      ||  !(th = fopen("pan.h", "w"))   /* header file   */
      ||  !(tt = fopen("pan.t", "w"))   /* transition matrix */
      ||  !(tm = fopen("pan.m", "w"))	/* forward  moves */
      ||  !(tb = fopen("pan.b", "w")))	/* backward moves */
    {	
      printf("spin: cannot create pan.[chtmb]\n");
      alldone(1);
    }

  plen_tab = (int *) emalloc((nrRdy + nrFun + 2) * sizeof(int));
  pmap_tab = (int **) emalloc((nrRdy + nrFun + 2) * sizeof(int *));
  memset(plen_tab, 0, (nrRdy + nrFun + 2) * sizeof(int));
  memset(pmap_tab, 0, (nrRdy + nrFun + 2) * sizeof(int *));

  clen_tab = (int *) emalloc((nrRdy + nrFun + 2) * sizeof(int));
  cmap_tab = (int **) emalloc((nrRdy + nrFun + 2) * sizeof(int *));
  memset(clen_tab, 0, (nrRdy + nrFun + 2) * sizeof(int));
  memset(cmap_tab, 0, (nrRdy + nrFun + 2) * sizeof(int *));

  fprintf(th, "#define Version  \"%s\"\n", Version);
  fprintf(th, "#define Source   \"%s\"\n\n", oFname->name + 3); /* trick */
  fprintf(th, "#define uchar    unsigned char\n");
  fprintf(th, "#define uint     unsigned int\n\n");
  fprintf(th, "#define WS       sizeof(long)\n");
  fprintf(th, "#define PAD(x)   ((WS - x %% WS) %% WS)\n\n");

  fprintf(th, "#if !defined(NOGC)&&!defined(GC0)&&!defined(GC1)\n");
  fprintf(th, "#ifndef NOSYMM\n");
  fprintf(th, "#define GC1 /* use symmetries only with mark and sweep */\n");
  fprintf(th, "#else\n");
  fprintf(th, "#define GC0 /* and reference counting otherwise */\n");
  fprintf(th, "#endif\n");
  fprintf(th, "#endif\n\n");

  fprintf(th, "#if !defined(NOGC)&&!defined(GC0)\n");
  fprintf(th, "#if !defined(NOSYMM)&&!defined(FULLSYMM)&&!defined(POSYMM)\n");
  fprintf(th, "#define POSYMM /* default symmetry reduction mode */\n");
  fprintf(th, "#endif\n");
  fprintf(th, "#else\n");
  fprintf(th, "#ifdef POSYMM\n");
  fprintf(th, "#undef POSYMM\n");
  fprintf(th, "#endif\n");
  fprintf(th, "#ifdef FULLSYMM\n");
  fprintf(th, "#undef FULLSYMM\n");
  fprintf(th, "#endif\n");
  fprintf(th, "#endif\n\n");

  if (Ntimeouts)
    fprintf(th, "#define NTIM	%d\n", Ntimeouts);

  if (Etimeouts)
    fprintf(th, "#define ETIM	%d\n", Etimeouts);

  if (has_remote)
    fprintf(th, "#define REM_REFS	%d\n", has_remote);

  if (has_last)
    fprintf(th, "#define HAS_LAST	%d\n", has_last);

  if (has_sorted)
    fprintf(th, "#define HAS_SORTED	%d\n", has_sorted);

  if (has_random)
    fprintf(th, "#define HAS_RANDOM	%d\n", has_random);
  
  if (!has_enabled && !eventmap)
    fprintf(th, "#define INLINE	1\n");

  if (has_enabled)
    fprintf(th, "#define HAS_ENABLED	1\n");

  if (has_unless)
    fprintf(th, "#define HAS_UNLESS	%d\n", has_unless);

  if (has_provided)
    fprintf(th, "#define HAS_PROVIDED	%d\n", has_provided);

  if (has_pcvalue)
    fprintf(th, "#define HAS_PCVALUE	%d\n", has_pcvalue);

  if (has_badelse)
    fprintf(th, "#define HAS_BADELSE	%d\n", has_badelse);

  if (has_np)
    fprintf(th, "#define HAS_NP	%d\n", has_np);

  fprintf(th, "#ifdef NP	/* includes np_ demon */\n");

  if (!has_np)
    fprintf(th, "#define HAS_NP	2\n");

  fprintf(th, "#define VERI	%d\n",	nrRdy);
  fprintf(th, "#define endclaim	3 /* none */\n");
  fprintf(th, "#endif\n");

  if (claimproc)
    {	
      claimnr = fproc(claimproc);
      /* NP overrides claimproc */
      fprintf(th, "#if !defined(NOCLAIM) && !defined NP\n");
      fprintf(th, "#define VERI	%d\n",	claimnr);
      fprintf(th, "#define endclaim	endstate%d\n",	claimnr);
      fprintf(th, "#endif\n");
    }

  genfunpars();

  if (eventmap)
    {	
      eventmapnr = fproc(eventmap);
      fprintf(th, "#define EVENT_TRACE	%d\n",	eventmapnr);
      fprintf(th, "#define endevent	endstate%d\n",	eventmapnr);

      if (eventmap[2] == 'o')	/* ":no_trace:" */
	fprintf(th, "#define NEGATED_TRACE	1\n");
    }

  fprintf(tc, "/*** Generated by %s ***/\n", Version);
  fprintf(tc, "/*** From source: %s ***/\n\n", oFname->name);

#ifdef __ALIGN
  fprintf(tc, "#define __ALIGN     1\n\n");
#endif

  ntimes(tc, 0, 1, Preamble);

  for (p = rdy; p; p = p->nxt)
    mst = max(p->s->maxel, mst);

  for (d = dfs; d; d = d->nxt)
    mst = max(d->body->maxel, mst);

  fprintf(tt, "#ifdef PEG\n");
  fprintf(tt, "struct T_SRC {\n");
  fprintf(tt, "  char *fl; int ln;\n");
  fprintf(tt, "} T_SRC[NTRANS];\n\n");

  fprintf(tt, "void\ntr_2_src(int m, char *file, int ln)\n");
  fprintf(tt, "{\n");
  fprintf(tt, "  T_SRC[m].fl = file;\n");
  fprintf(tt, "  T_SRC[m].ln = ln;\n");
  fprintf(tt, "}\n\n");

  fprintf(tt, "void\nputpeg(int n, int m)\n{\n");
  fprintf(tt, "  printf(\"%%5d\ttrans %%4d \", m, n);\n");
  fprintf(tt, "  printf(\"file %%s line %%3d\\n\",\n");
  fprintf(tt, "          T_SRC[n].fl, T_SRC[n].ln);\n");
  fprintf(tt, "}\n");

  fprintf(tt, "#else\n");
  fprintf(tt, "#define tr_2_src(m,f,l)\n");
  fprintf(tt, "#endif /* PEG */\n\n");

  fprintf(tt, "void settable(void)\n{\n  Trans *T;\n");
  fprintf(tt, "  Trans *settr(int, int, int, int, int,");
  fprintf(tt, " char *, int, int, int);\n\n");
  fprintf(tt, "  trans = (Trans ***) ");
  fprintf(tt, "emalloc(%d * sizeof(Trans **));\n", nrRdy + nrFun + 1);
			/* +1 for np_ automaton */

  fprintf(tm, "  switch (t->forw) {\n");
  fprintf(tm, "    default: Uerror(\"bad forward move\");\n\n");

  fprintf(tb, "	switch (t->back) {\n");
  fprintf(tb, "	  default: Uerror(\"bad return move\");\n\n");
  fprintf(tb, "	  case  0: goto R999; /* nothing to undo */\n\n");

  dfa_all(rdy, dfs);

  heap_init_size = global_assign();

  for (p = rdy; p; p = p->nxt) {
    Proc = p;
    putproc(p);
  }

  Proc = 0;

  tm_predef_np();
  tt_predef_np();

  genpushpars();

  for (d = dfs; d; d = d->nxt)
    putfun(d);

  fprintf(tt, "}\n\n");	/* end of settable() */
  fprintf(tm, "}\n\n");
  fprintf(tb, "}\n\n");

  ntimes(tt, 0, 1, Tail);

  genheader();
  genaddproc();
  genaddheap();
  genother();

  /*
   * genaddqueue() and genunio() use the queue list `qtab' 
   * built up while generating global and local initializers.
   */

  genaddqueue();
  genunio();

  genconditionals();

  if (!run) 
    fatal("no runable process", (char *)0);

  fprintf(tc, "void active_procs(void)\n{\n");
  reverse_procs(run);
  fprintf(tc, "}\n\n");

  fprintf(tc, "#include <spin.c>\n");
  fprintf(th, "#define NTRANS\t\t%d\n\n", uniq);
  fprintf(th, "#ifdef PEG\n");
  fprintf(th, "long peg[NTRANS];\n");
  fprintf(th, "#endif\n\n");

  fprintf(th, "int p_len[] = {\n\t");
  for (i = 0; i < nrRdy + nrFun + 2; i ++) {
    fprintf(th, "%d, ", plen_tab[i]);
    len_max = max(len_max, plen_tab[i]);
  }
  fprintf(th, "\n};\n\n");

  fprintf(th, "int p_map[][%d] = {\n", len_max);
  for (i = 0; i < nrRdy + nrFun + 2; i ++) {
    fprintf(th, "\t{");
    for (j = 0; j < plen_tab[i]; j ++)
      fprintf(th, "%d, ", pmap_tab[i][j]);
    fprintf(th, "},\n");
  }
  fprintf(th, "\n};\n\n");

  len_max = 0;
  fprintf(th, "int c_len[] = {\n\t");
  for (i = 0; i < nrRdy + nrFun + 2; i ++) {
    fprintf(th, "%d, ", clen_tab[i]);
    len_max = max(len_max, clen_tab[i]);
  }
  fprintf(th, "\n};\n\n");

  fprintf(th, "int c_map[][%d] = {\n", len_max);
  for (i = 0; i < nrRdy + nrFun + 2; i ++) {
    fprintf(th, "\t{");
    for (j = 0; j < clen_tab[i]; j ++)
      fprintf(th, "%d, ", cmap_tab[i][j]);
    fprintf(th, "},\n");
  }
  fprintf(th, "\n};\n\n");

  fprintf(th, "#include <spin.h>\n");

  fclose(tc);
  fclose(th);
  fclose(tt);
  fclose(tm);
  fclose(tb);
}


static struct AA {
  char TT[9];	char CC[8];
} _BB[4] = {
  { "Q_FULL_F",	"q_full" },
  { "Q_FULL_T",	"!q_full" },
  { "Q_EMPT_F",	"!q_len" },
  { "Q_EMPT_T",	"q_len" }
};


static void
Done_case(char *nm, Symbol *z, int base)
{	
  int j, k;
  int nid = z->Nid;
  int qln = z->nel;

  fprintf(tc, "  case %d: if (", nid);

  for (j = 0; j < 4; j++)
    {	
      fprintf(tc, "(t->ty[i] == %s && (", _BB[j].TT);
      for (k = 0; k < qln; k++)
	{	
	  if (k > 0)
	    fprintf(tc, "\n              || ");

	  fprintf(tc, "%s(%s(%d, %s))", _BB[j].CC, nm, 
		  base + k * (z->bitsize/8), typ2c(z));
	}
	
      fprintf(tc, "))\n              ");
      
      if (j < 3)
	fprintf(tc, "|| ");
      else
	fprintf(tc, "   ");
    }

  fprintf(tc, ") return 0; break;\n\n");
}


static void
walk2_struct(char *pre, Symbol *s, int base)
{
  Lextok *fp, *tl;
  int ix, newb;

  ini_struct(s);

  for (ix = 0; ix < s->nel; ix ++)
    for (fp = s->Sval[ix + 1]; fp; fp = fp->rgt)
      for (tl = fp->lft; tl; tl = tl->rgt)
	{
	  newb = base + ix * s->bitsize / 8 + tl->sym->offset;

	  switch (tl->sym->type)
	    {
	    case STRUCT:
	      walk2_struct(pre, tl->sym, newb);
	      break;

	    case CHAN:
	      Done_case(pre, tl->sym, newb);
	    }
	}
}


static void
dolen(Symbol *s, char *pre, int ai, int qln)
{
  if (ai > 0)
    fprintf(tc, "\n              || ");

  fprintf(tc, "%s(", pre);

  if (!(s->hidden & 1))
    {	
      if (s->context)
	fprintf(tc, "ll(");
      else
	fprintf(tc, "gg(");

      fprintf(tc, "%d, %s)", s->offset + ai * (s->bitsize/8), typ2c(s));
    }
  else
    fprintf(tc, "%s", s->name);

  fprintf(tc, ")");
}


static void
Docase(Symbol *s, int nid)
{	
  int i, j;

  fprintf(tc, "  case %d: if (", nid);
  for (j = 0; j < 4; j++)
    {	
      fprintf(tc, "(t->ty[i] == %s && (", _BB[j].TT);

      for (i = 0; i < s->nel; i++)
	dolen(s, _BB[j].CC, i, s->nel);

      fprintf(tc, "))\n              ");

      if (j < 3)
	fprintf(tc, "|| ");
      else
	fprintf(tc, "   ");
    }
  
  fprintf(tc, ") return 0; break;\n\n");
}


static void
genconditionals(void)
{	
  Symbol *s;
  int last = 0;
  Ordered *walk;

  extern Ordered *all_names;

  fprintf(tc, "int q_cond(short II, Trans *t)\n");
  fprintf(tc, "{\n");
  fprintf(tc, "  int i = 0;\n\n");
  fprintf(tc, "  for (i = 0; i < 6; i ++) {\n");
  fprintf(tc, "    if (t->ty[i] == TIMEOUT_F)\n");
  fprintf(tc, "      return %s;\n\n", (Etimeouts) ? "(!(trpt->tau&1))" : "1");

  fprintf(tc, "  if (t->ty[i] == ALPHA_F)\n");
  fprintf(tc, "#ifdef GLOB_ALPHA\n");
  fprintf(tc, "    return 0;\n");
  fprintf(tc, "#else\n");
  fprintf(tc, "    return");
  fprintf(tc, " (II + 1 == (short) now._nr_pr && II + 1 < MAXPROC);\n");
  fprintf(tc, "#endif /* GLOB_ALPHA */\n\n");

  /* we switch on the chan name from the spec (as identified by
   * the corresponding Nid number) rather than the actual qid
   * because we cannot predict at compile time which specific qid
   * will be accessed by the statement at runtime.  that is:
   * we do not know which qid to pass to q_cond at runtime
   * but we do know which name is used.  if it's a chan array, we
   * must check all elements of the array for compliance (bummer)
   */

  fprintf(tc, "  switch (t->qu[i]) {\n");
  fprintf(tc, "  case 0: break;\n");

  for (walk = all_names; walk; walk = walk->next)
    {	
      if (!(s = walk->entry))
	continue;

      if (s->owner) 
	continue;

      if (s->type == CHAN) 
	{
	  if (last == s->Nid) 
	    continue;	/* chan array */

	  last = s->Nid;
	  Docase(s, last);
	} 
      else 
	/* struct may contain a chan */
	if (s->type == STRUCT) 
	  {
	    char pre[128];

	    if (!(s->hidden & 1)) 
	      {
		if (s->context)
		  sprintf(pre, "ll");
		else
		  sprintf(pre, "gg");
		
		walk2_struct(pre, s, s->offset);
	      }
	    else
	      fatal("cannot happen (genconditionals)", (char *)0);
	  }
    }
	
  fprintf(tc, "  default: Uerror(\"unknown qid - q_cond\");\n");
  fprintf(tc, "           return 0;\n");
  fprintf(tc, "    }\n");
  fprintf(tc, "  }\n");
  fprintf(tc, "	 return 1;\n");
  fprintf(tc, "}\n\n\n");
}


static void 
putproc(ProcList *p)
{
  int isz = 0;

  isz += doformal(0, ASSIGN, p->s, isz, 0);
  isz += dolocal(0, ASSIGN, p->s, isz);

  __align(isz);

  p->s->sz = isz;
  p->s->bs = isz;

  Pid = p->tn;
  Det = p->det;

  AllGlobal = (p->prov) ? 1 : 0; /* process has provided clause */

  fprintf(th, "\nshort nstates%d = %d;\t/* %s */\n",
	  p->tn, p->s->maxel, p->n->name);

  if (p->tn == claimnr)
    fprintf(th, "#define nstates_claim	nstates%d\n", p->tn);

  if (p->tn == eventmapnr)
    fprintf(th, "#define nstates_event	nstates%d\n", p->tn);

  fprintf(th, "#define endstate%d\t%d\n\n",
	  p->tn, p->s->last->seqno);

  fprintf(tm, "\n\t\t/* PROC %s */\n", p->n->name);
  fprintf(tb, "\n\t\t/* PROC %s */\n", p->n->name);

  fprintf(tt, "\n\t\t/* proctype %d: %s */\n", p->tn, p->n->name);
  fprintf(tt, "\n  trans[%d] = (Trans **)", p->tn);
  fprintf(tt, " emalloc(%d * sizeof(Trans *));\n\n", p->s->maxel);

  if (Pid == eventmapnr)
    {	
      fprintf(th, "\n#define in_s_scope(x_y3_)\t0");
      fprintf(tc, "\n#define in_r_scope(x_y3_)\t0");
    }
  
  enter_seq(p->s);
  put_seq(p->s, 2, 0);
  exit_seq();

  if (Pid == eventmapnr)
    {	
      fprintf(th, "\n\n");
      fprintf(tc, "\n\n");
    }

  dumpsrc(p->s->maxel, p->tn);
}


static void
addTpe(int x)
{	int i;

	if (x <= 2) return;

	for (i = 0; i < T_sum; i++)
		if (TPE[i] == x)
			return;
	TPE[(T_sum++)%2] = x;
}

static void
cnt_seq(Sequence *s)
{	Element *f;
	SeqList *h;

	if (s)
	for (f = s->frst; ; f = f->nxt)
	{	Tpe(f->n);	/* sets EPT */
		addTpe(EPT[0]);
		addTpe(EPT[1]);
		for (h = f->sub; h; h = h->nxt)
			cnt_seq(h->this);
		if (f == s->last)
			break;
	}
}

static void
typ_seq(Sequence *s)
{
	T_sum = 0;
	TPE[0] = 2; TPE[1] = 0;
	cnt_seq(s);
	if (T_sum > 2)		/* more than one type */
	{	TPE[0] = 5*DELTA;	/* non-mixing */
		TPE[1] = 0;
	}
}

static int
hidden(Lextok *n)
{       
  if (n)
    switch (n->ntyp) {
    case  FULL: case  EMPTY:
    case NFULL: case NEMPTY: case TIMEOUT:         
      Nn[(T_mus++)%2] = n;
      break;
      
    case '!': case UMIN: case '~': case ASSERT: case 'c':
      (void) hidden(n->lft);
      break;

    case '/': case '*': case '-': case '+':
    case '%': case LT:  case GT: case '&': case '^':
    case '|': case LE:  case GE:  case NE: case '?':
    case EQ:  case OR:  case AND: case LSHIFT: case RSHIFT:
      (void) hidden(n->lft);
      (void) hidden(n->rgt);
      break;
    }
  
  return T_mus;
}

static int
getNid(Lextok *n)
{
  if (n->sym->type & STRUCT
      &&  n->rgt && n->rgt->lft)
    return getNid(n->rgt->lft);

  if (n->sym->Nid == 0)
    fatal("bad channel name '%s'", (n->sym)?n->sym->name:"no name");
	
  return n->sym->Nid;
}

static int
valTpe(Lextok *n)
{	int res = 2;
	/*
	2 = local
	2+1	    .. 2+1*DELTA = nfull,  's'	- require q_full==false
	2+1+1*DELTA .. 2+2*DELTA = nempty, 'r'	- require q_len!=0
	2+1+2*DELTA .. 2+3*DELTA = empty	- require q_len==0
	2+1+3*DELTA .. 2+4*DELTA = full		- require q_full==true
	5*DELTA = non-mixing (i.e., always makes the selection global)
	6*DELTA = timeout (conditionally safe)
	7*DELTA = @, process deletion (conditionally safe)
	 */
	switch (n->ntyp) { /* a series of fall-thru cases: */
	case   FULL:	res += DELTA;		/* add 3*DELTA + chan nr */
	case  EMPTY:	res += DELTA;		/* add 2*DELTA + chan nr */
	case    'r':
	case NEMPTY:	res += DELTA;		/* add 1*DELTA + chan nr */
	case    's':
	case  NFULL:	res += getNid(n->lft);	/* add channel nr */
			break;

	case TIMEOUT:	res = 6*DELTA; break;
	case STOP:      res = 7*DELTA; break;
	default:	break;
	}
	return res;
}

static void
Tpe(Lextok *n)	/* mixing in selections */
{
	EPT[0] = 2; EPT[1] = 0;

	if (!n) return;

	T_mus = 0;
	Nn[0] = Nn[1] = ZN;

	if (n->ntyp == 'c')
	{	if (hidden(n->lft) > 2)
		{	EPT[0] = 5*DELTA; /* non-mixing */
			EPT[1] = 0;
			return;
		}
	} else
		Nn[0] = n;

	if (Nn[0]) EPT[0] = valTpe(Nn[0]);
	if (Nn[1]) EPT[1] = valTpe(Nn[1]);
}

static void
put_escp(Element *e)
{	
  int n;
  SeqList *x;

  if (e->esc && e->n->ntyp != GOTO && e->n->ntyp != '.')
    {	
      for (x = e->esc, n = 0; x; x = x->nxt, n++)
	{	
	  int i = huntele(x->this->frst, e->status)->seqno;
	  fprintf(tt, "\ttrans[%d][%d]->escp[%d] = %d;\n",
		  Pid, e->seqno, n, i);
	  fprintf(tt, "\treached%d[%d] = 1;\n",
		  Pid, i);
	}
		
      for (x = e->esc, n=0; x; x = x->nxt, n++)
	{	
	  fprintf(tt, "	/* escape #%d: %d */\n", n,
		  huntele(x->this->frst, e->status)->seqno);
	  put_seq(x->this, 2, 0);	/* args?? */
	}
      
      fprintf(tt, "	/* end-escapes */\n");
    }
}


static void
put_sub(Element *e, int Tt0, int Tt1)
{	
  Sequence *s = e->n->sl->this;
  Element *g = ZE;
  int a;

  patch_atomic(s);
  putskip(s->frst->seqno);
  g = huntstart(s->frst);
  a = g->seqno;

  if ((e->n->ntyp == ATOMIC
       ||  e->n->ntyp == D_STEP)
      &&  scan_seq(s))
    mark_seq(s);
	
  s->last->nxt = e->nxt;

  typ_seq(s);	/* sets TPE */

  if (e->n->ntyp == D_STEP)
    {	
      int inherit = (e->status&(ATOM|L_ATOM));
      fprintf(tm, "\tcase  %d:", uniq++);

      fprintf(tm, "/* STATE %d - line %d %s - [",
	      e->seqno, e->n->ln, e->n->fn->name);
      comment(tm, e->n, 0);
      fprintf(tm, "] */\n\t\t");

      if (s->last->n->ntyp == BREAK)
	OkBreak = target(huntele(s->last->nxt, s->last->status))->Seqno;
      else
	OkBreak = -1;

      if (!putcode(tm, s, e->nxt, 0, e->n->ln))
	fprintf(tm, "\t\tm = %d; goto P999;\n\n", getweight(s->frst->n));
	
      fprintf(tb, "\tcase  %d: ", uniq-1);
      fprintf(tb, "/* STATE %d */\n", e->seqno);
      fprintf(tb, "\t\tsv_restor();\n");
      fprintf(tb, "\t\tgoto R999;\n");

      if (e->nxt)
	a = huntele(e->nxt, e->status)->seqno;
      else
	a = 0;
	
      tr_map(uniq-1, e);
      fprintf(tt, "/*->*/\ttrans[%d][%d]\t= ",
	      Pid, e->seqno);
		
      fprintf(tt, "settr(%d,%d,%d,%d,%d,\"",
	      e->Seqno, D_ATOM|inherit, a, uniq-1, uniq-1);
      comment(tt, e->n, e->seqno);
      fprintf(tt, "\", %d, ", (s->frst->status & I_GLOB) ? 1 : 0);
      fprintf(tt, "%d, %d);\n", TPE[0], TPE[1]);
      put_escp(e);
    } 
  else {	
    /* ATOMIC or NON_ATOMIC */
    fprintf(tt, "\tT = trans[ %d][%d] = ", Pid, e->seqno);
    fprintf(tt, "settr(%d,%d,0,0,0,\"",
	    e->Seqno, (e->n->ntyp == ATOMIC)?ATOM:0);

    comment(tt, e->n, e->seqno);

    if ((e->status&CHECK2)
	||  (g->status&CHECK2))
      s->frst->status |= I_GLOB;

    fprintf(tt, "\", %d, %d, %d);",
	    (s->frst->status & I_GLOB) ? 1 : 0, Tt0, Tt1);
    
    blurb(tt, e);
    fprintf(tt, "\tT->nxt\t= ");
    fprintf(tt, "settr(%d,%d,%d,0,0,\"",
	    e->Seqno, (e->n->ntyp == ATOMIC) ? ATOM : 0, a);
		
    comment(tt, e->n, e->seqno);

    fprintf(tt, "\", %d, ", (s->frst->status & I_GLOB)?1:0);

    if (e->n->ntyp == NON_ATOMIC) {	
      fprintf(tt, "%d, %d);", Tt0, Tt1);
      blurb(tt, e);
      put_seq(s, Tt0, Tt1);
    } 
    else {	
      fprintf(tt, "%d, %d);", TPE[0], TPE[1]);
      blurb(tt, e);
      put_seq(s, TPE[0], TPE[1]);
    }
  }
}


static void
put_fb(Sequence *s)
{
  int sz = 0;

  if (s->cnxt)
    fprintf(tm, "/* --> %s() */\n\t\t", s->cnxt->name);

  enter_seq(s);

  sz += doformal(0, ASSIGN, s, sz, 0);
  sz += dolocal(0, ASSIGN, s, sz);

  __align(sz);

  s->sz = sz;
  s->bs = sz;

  if (sz) {
    fprintf(tm, "push(II, %ld);\n\t\t", sz);
    fprintf(tm, "/* locals */\n\t\t");
    fprintf(tm, "{ int h = II, _t = Q_LOCAL;\n");
    fprintf(tm, "#ifdef POSYMM\n\t\t");
    fprintf(tm, "  scalar_t *_s = (scalar_t *) 0;\n");
    fprintf(tm, "#endif /* POSYMM */\n\n");
    dolocal(tm, INITVAR, s, 0);
    fprintf(tm, "\t\t}\n");

    fprintf(tm, "\n#ifdef GC0\n");
    dolocal(tm, GC0_INCR, s, 0);
    fprintf(tm, "#endif\n\t\t");

    doformal(0, GC1_MAP, s, 0, 0);
    doformal(0, CH_MAP, s, 0, 0);
    dolocal(0, GC1_MAP, s, 0);
    dolocal(0, CH_MAP, s, 0);
  }
}


static void
put_fe(Sequence *s)
{
  if (s->cnxt)
    fprintf(tm, "/* <-- %s() */\n\t\t", s->cnxt->name);

  fprintf(tm, "sv_save(&now);\n\t\t");

  if (s->sz) {
    fprintf(tm, "\n#ifdef GC0\n");
    doformal(tm, GC0_NOSV, s, 0, 0);
    dolocal(tm, GC0_NOSV, s, 0);
    fprintf(tm, "#endif /* GC0 */\n\t\t");
  }

  fprintf(tm, "ret(II, 0, 0);\n\t\t");
  fprintf(tm, "rt[II] = 1;\n\t\t");
  fprintf(tm, "vt[II] = 1;\n\t\t");

  if (s->sz) {
    fprintf(tm, "\n#ifdef GC1\n\t\t");
    fprintf(tm, "gc_scan();\n\t\t");
    fprintf(tm, "gc_sweep(0);\n");
    fprintf(tm, "#endif /* GC1 */\n\t\t");
  }
}



static void
undo_fb(Sequence *s)
{
  if (s->sz) {
    fprintf(tb, "\n#ifdef GC0\n");
    dolocal(tb, GC0_INC_UNDO, s, 0);
    fprintf(tb, "#endif\n\t\t");
    fprintf(tb, "pop(II, %ld);\n\t\t", s->sz);
  }
}


static void
undo_fe(Sequence *s)
{
  fprintf(tb, "sv_restor();\n\t\t");
}


static void
put_bb(SeqList *sl)
{
  Sequence *s; 
  Sequence *i; 

  if (!sl)
    return;

  s = sl->this;
  i = cur_s->this;

  if (s->cnxt)
    fprintf(tm, "/* --> %s */\n\t\t", s->cnxt->name);
  
  enter_seq(s);
  
  s->sz = dolocal(0, ASSIGN, s, i->bs);

  __align(s->sz);

  s->bs = i->bs + s->sz;
  
  if (s->sz) {
    fprintf(tm, "push(II, %ld);\n\t\t", s->sz);	
    fprintf(tm, "/* locals */\n\t\t");
    fprintf(tm, "{ int h = II, _t = Q_LOCAL;\n");
    fprintf(tm, "#ifdef POSYMM\n");
    fprintf(tm, "  scalar_t *_s = (scalar_t *) 0;\n");
    fprintf(tm, "#endif /* POSYMM */\n\n");
    dolocal(tm, INITVAR, s, 0);
    fprintf(tm, "\t\t}\n");

    fprintf(tm, "\n#ifdef GC0\n");
    dolocal(tm, GC0_INCR, s, 0);
    fprintf(tm, "#endif\n\t\t");
    
    dolocal(0, GC1_MAP, s, 0);
    dolocal(0, CH_MAP, s, 0);
  }
  
  fprintf(tm, "\n\t\t");

  put_bb(sl->nxt);
}


static void
put_be(SeqList *sl)
{
  if (!sl)
    return;

  put_be(sl->nxt);

  if (sl->this->cnxt)
    fprintf(tm, "/* <-- %s */\n\t\t", sl->this->cnxt->name);
      
  if (sl->this->sz) {
    fprintf(tm, "\n#ifdef GC0\n");
    dolocal(tm, GC0_SAVE, sl->this, 0);
    fprintf(tm, "#endif\n\t\t");

    fprintf(tm, "pop(II, %ld);\n\t\t", sl->this->sz);

    fprintf(tm, "\n#ifdef GC1\n\t\t");
    fprintf(tm, "gc_scan();\n\t\t");
    fprintf(tm, "gc_sweep(1);\n");
    fprintf(tm, "#endif /* GC1 */\n\t\t");
  }

  exit_seq();
}


static void
put_jmp(SeqList *sl)
{
  if (!sl)
    return;

  put_jmp(sl->nxt);

  if (sl->this->cnxt)
    fprintf(tm, "/* <-- %s */\n\t\t", sl->this->cnxt->name);
      
  if (sl->this->sz) {
    fprintf(tm, "\n#ifdef GC0\n");
    dolocal(tm, GC0_SAVE, sl->this, 0);
    fprintf(tm, "#endif\n\t\t");
    fprintf(tm, "pop(II, %ld);\n\t\t", sl->this->sz); 
  }
}


static void
undo_bb(SeqList *sl)
{
  if (!sl)
    return;

  undo_bb(sl->nxt);

  if (sl->this->sz) {
    fprintf(tb, "pop(II, %ld);\n\t\t", sl->this->sz);
    fprintf(tb, "\n#ifdef GC0\n");
    dolocal(tb, GC0_INC_UNDO, sl->this, 0);
    fprintf(tb, "#endif\n\t\t");
  }
}


static void
undo_be(SeqList *sl)
{
  if (!sl) 
    return;

  if (sl->this->sz) {
    fprintf(tb, "push(II, %ld);\n\t\t", sl->this->sz);
    fprintf(tb, "\n#ifdef GC0\n");
    dolocal(tb, GC0_DEC_UNDO, sl->this, 0);
    fprintf(tb, "#endif\n\t\t");
  }

  undo_be(sl->nxt);
}


static void
undo_jmp(SeqList *sl)
{
  if (!sl) 
    return;

  if (sl->this->sz) {
    fprintf(tb, "push(II, %ld);\n\t\t", sl->this->sz);
    fprintf(tb, "\n#ifdef GC0\n");
    dolocal(tb, GC0_DEC_UNDO, sl->this, 0);
    fprintf(tb, "#endif\n\t\t");
  }

  undo_be(sl->nxt);
}


static void
put_el(Element *e, int Tt0, int Tt1)
{	
  int n, a, Global_ref;
  Element *g = ZE;

#if (DEBUG >= 1)
  SeqList *x;
  printf("put_el %d (%d) (->%d (%d)) (", e->Seqno, e->seqno,
	 (e->nxt)?e->nxt->Seqno:-1, (e->nxt)?e->nxt->seqno:-1);
  comment(stdout, e->n, 0);
  printf(")");
  if (e->esc) printf(" - escaped by ");
  for (x = e->esc; x; x = x->nxt)
    {	
      printf("%d (", x->this->frst->Seqno);
      comment(stdout, x->this->frst->n, 0);
      printf(") ");
    }
	
  printf("\n");
#endif
  
  if (e->n->ntyp == GOTO)
    {	
      g = get_lab(e->n, 1);
      g = huntele(g, e->status);
      cross_dsteps(e->n, g->n);
      a = g->seqno;
    } 
  else 
    if (e->nxt)
      {	
	g = huntele(e->nxt, e->status);
	a = g->seqno;
      } else
	a = 0;
	
  if (g && (g->status & CHECK2	        /* entering remotely ref'd state */
	    || e->status & CHECK2))	/* leaving  remotely ref'd state */
    e->status |= I_GLOB;

  tr_map(uniq, e);
  fprintf(tt, "\ttrans[%d][%d]\t= ", Pid, e->seqno);

  fprintf(tm, "\tcase  %d: /* STATE %d - line %d %s - [",
	  uniq ++, e->seqno, e->n->ln, e->n->fn->name);

  comment(tm, e->n, 0);
  fprintf(tm, "] */\n\t\t");

  if (e->n->ntyp != 'r' && Pid != claimnr)
    fprintf(tm, "IfNotBlocked\n\t\t");

  Global_ref = (e->status & I_GLOB) ? 1 : has_global(e, e->n);

  if (e->tag & FB)
    put_fb(e->fsq);

  if (e->tag & BB)
    put_bb(e->sbl);

  putstmnt(tm, e->n, e->seqno);

  if (e->n->ntyp != ELSE && Det)
    {	
      fprintf(tm, ";\n\t\tif (trpt->o_pm&1)\n\t\t");
      fprintf(tm, "\tuerror(\"non-determinism in D_proctype\")");
    }

  if (any_undo(e->n)
      || e->tag & FB
      || e->tag & FE
      || e->tag & BE
      || e->tag & BB
      || e->tag & JMP)
    {
      fprintf(tb, "\tcase  %d: ", uniq-1);
      fprintf(tb, "/* STATE %d */\n\t\t", e->seqno);

      if (e->n->ntyp != RETURN
	  && e->n->ntyp != RETFUN)
	{
	  if (e->tag & FE)
	    undo_fe(e->fsq);

	  if (e->tag & BE
	      || e->tag & JMP) {
	    undo_be(e->sel);

	    fprintf(tb, "\n#ifdef GC1\n\t\t");
	    fprintf(tb, "if (trpt->gc) sv_restor();\n");
	    fprintf(tb, "#endif\n");
	  }
	}

      if (any_undo(e->n))
	{
	  _isok++;
	  undostmnt(e->n, e->seqno);
	  _isok--;
	}

      fprintf(tb, ";\n\t\t"); 

      if (e->tag & BB)
	undo_bb(e->sbl);

      if (e->tag & FB)
	undo_fb(e->fsq);

      fprintf(tb, "goto R999;\n");
      fprintf(tt, "settr(%d,%d,%d,%d,%d,\"",
	      e->Seqno, e->status&2, a, uniq-1, uniq-1);
    } else
      fprintf(tt, "settr(%d,%d,%d,%d,0,\"",
	      e->Seqno, e->status&2, a, uniq-1);

  fprintf(tm, ";\n\t\t");

  if (e->n->ntyp != RETURN
      && e->n->ntyp != RETFUN)
    {
      if (e->tag & BE)
	put_be(e->sel);
      
      if (e->tag & JMP)
	put_jmp(e->sel);

      if (e->tag & BE
	  || e->tag & JMP)
	{
	  fprintf(tm, "\n#ifdef GC1\n\t\t");
	  fprintf(tm, "gc_scan();\n\t\t");
	  fprintf(tm, "gc_sweep(1);\n");
	  fprintf(tm, "#endif\n\t\t");
	}

      if (e->tag & FE)
	put_fe(e->fsq);	
    }

  n = getweight(e->n);
  fprintf(tm, "m = %d; goto P999;\n", n);

  comment(tt, e->n, e->seqno);
  fprintf(tt, "\", %d, ", Global_ref);

  if (Tt0 != 2)
      fprintf(tt, "%d, %d);\n", Tt0, Tt1);
  else { 
    Tpe(e->n);	/* sets EPT */
    fprintf(tt, "%d, %d);\n", EPT[0], EPT[1]);
  }
  
  put_escp(e);
}


static void
nested_unless(Element *e, Element *g)
{	struct SeqList *y = e->esc, *z = g->esc;

	for ( ; y && z; y = y->nxt, z = z->nxt)
		if (z->this != y->this)
			break;
	if (!y && !z)
		return;

	if (g->n->ntyp != GOTO
	&&  g->n->ntyp != '.'
	&&  e->sub->nxt)
	{	printf("error: (%s:%d) saw 'unless' on a guard:\n",
			(e->n)?e->n->fn->name:"-",
			(e->n)?e->n->ln:0);
		printf("=====>instead of\n");
		printf("	do (or if)\n");
		printf("	:: ...\n");
		printf("	:: stmnt1 unless stmnt2\n");
		printf("	od (of fi)\n");
		printf("=====>use\n");
		printf("	do (or if)\n");
		printf("	:: ...\n");
		printf("	:: stmnt1\n");
		printf("	od (or fi) unless stmnt2\n");
		printf("=====>or rewrite\n");
	}
}


void 
put_seq(Sequence *s, int Tt0, int Tt1)
{	
  SeqList *h;
  Element *e, *g;
  int a;

#if (DEBUG >= 1)
  printf("put_seq %d - %d [%d] (%d,%d)\n",
	 s->frst->Seqno,
	 s->last->Seqno,
	 s->extent->Seqno,
	 Tt0, Tt1);
#endif

  for (e = s->frst; e; e = e->nxt)
    {
#if (DEBUG >= 1)
      printf("put_seq %d (%d) (; %d (%d)) ",
	     e->Seqno, e->seqno,
	     e->nxt ? e->nxt->Seqno:-1,
	     e->nxt ? e->nxt->seqno:-1);
      comment(stdout, e->n, 0);
      printf(" (%s)\n", (e->status & DONE) ? "done" : "new");
#endif
      if (e->status & DONE) 
	goto checklast;	/* was continue */

      e->status |= DONE;

      if (e->n->ln)
	putsrc(e->n->ln, e->seqno);

      if (e->n->ntyp == UNLESS)
	{
#if (DEBUG >= 1)
	  printf("\t(unless) %d\n",
		 e->sub->this->frst->seqno);
#endif
	  put_seq(e->sub->this, Tt0, Tt1);
	} 
      else 
	if (e->sub)
	  {	
	    fprintf(tt, "\tT = trans[%d][%d] = ", Pid, e->seqno);
	    fprintf(tt, "settr(%d, %d, 0, 0, 0,\"", e->Seqno, e->status&2);
	    comment(tt, e->n, e->seqno);
	      
	    if (e->status & CHECK2)
	      e->status |= I_GLOB;
	    
	    fprintf(tt, "\", %d, %d, %d);", 
		    (e->status & I_GLOB) ? 1 : 0, Tt0, Tt1);
	    blurb(tt, e);
	      
	    for (h = e->sub; h; h = h->nxt)
	      {	
		putskip(h->this->frst->seqno);
		g = huntstart(h->this->frst);
		
		if (g->esc)
		  nested_unless(e, g);
		  
		a = g->seqno;
		
		if (h->nxt)
		  fprintf(tt, "\tT = T->nxt\t= ");
		else
		  fprintf(tt, "\t    T->nxt\t= ");
		
		fprintf(tt, "settr(%d, %d, %d, 0, 0,\"",
			e->Seqno, e->status&2, a);

		comment(tt, e->n, e->seqno);

		if (g->status & CHECK2)
		  h->this->frst->status |= I_GLOB;
		
		fprintf(tt, "\", %d, %d, %d);",
			(h->this->frst->status & I_GLOB) ? 1 : 0, Tt0, Tt1);

		blurb(tt, e);
	      }
#if (DEBUG >= 1)
	    printf(" subsequence\n");
#endif
	    for (h = e->sub; h; h = h->nxt)
	      put_seq(h->this, Tt0, Tt1);
	  } 
	else {
	  if (e->n->ntyp == ATOMIC
	      ||  e->n->ntyp == D_STEP
	      ||  e->n->ntyp == NON_ATOMIC)
	    {
#if (DEBUG >= 1)
	      printf(" putsub\n");
#endif
	      put_sub(e, Tt0, Tt1);
	    }
	  else {
#if (DEBUG >= 1)
	    printf(" putel\n");
#endif
	    put_el(e, Tt0, Tt1);
	  }
	}
    checklast:
      if (e == s->last)
	{
#if (DEBUG >= 1)
	  printf(" islast\n");
#endif
	  break;
	}
    }

#if (DEBUG >= 1)
  printf(" putseq returns\n");
#endif
}


static void
patch_atomic(Sequence *s)	/* catch goto's that break the chain */
{	Element *f, *g;
	SeqList *h;

	for (f = s->frst; ; f = f->nxt)
	{	if (f->n && f->n->ntyp == GOTO)
		{	g = get_lab(f->n,1);
			cross_dsteps(f->n, g->n);
			if ((f->status & (ATOM|L_ATOM))
			&& !(g->status & (ATOM|L_ATOM)))
			{	f->status &= ~ATOM;
				f->status |= L_ATOM;
			}
		} else
		for (h = f->sub; h; h = h->nxt)
			patch_atomic(h->this);
		if (f == s->extent)
			break;
	}
}

static void
mark_seq(Sequence *s)
{	
  Element *f;
  SeqList *h;

  for (f = s->frst; ; f = f->nxt)
    {	
      f->status |= I_GLOB;
      
      for (h = f->sub; h; h = h->nxt)
	mark_seq(h->this);
	
      if (f == s->last)
	return;
    }
}

static Element *
find_target(Element *e)
{	Element *f;

	if (!e) return e;

	if (t_cyc++ > 32)
	{	fatal("cycle of goto jumps", (char *) 0);
	}
	switch (e->n->ntyp) {
	case  GOTO:
		f = get_lab(e->n,1);
		cross_dsteps(e->n, f->n);
		f = find_target(f);
		break;
	case BREAK:
		if (e->nxt)
		f = find_target(huntele(e->nxt, e->status));
		/* else fall through */
	default:
		f = e;
		break;
	}
	return f;
}

Element *
target(Element *e)
{
	if (!e) return e;
	lineno = e->n->ln;
	Fname  = e->n->fn;
	t_cyc = 0;
	return find_target(e);
}

static int
scan_seq(Sequence *s)
{	
  Element *f, *g;
  SeqList *h;

  for (f = s->frst; ; f = f->nxt)
    {	if ((f->status&CHECK2)
	    ||  has_global(f, f->n))
      return 1;
	
    if (f->n->ntyp == GOTO)	/* may reach other atomic */
      {	g = target(f);
      if (g
	  && !(f->status & L_ATOM)
	  && !(g->status & (ATOM|L_ATOM)))
	{	
	  fprintf(tt, "	/* goto mark-down, ");
	  fprintf(tt, "line %d - %d */\n",
		  f->n->ln, (g->n)?g->n->ln:0);
	  return 1; /* assume worst case */
	}	
      }

    for (h = f->sub; h; h = h->nxt)
      if (scan_seq(h->this))
	return 1;

    if (f == s->last)
      break;
    }

  return 0;
}


static int
glob_args(Element *e, Lextok *n)
{	
  int result = 0;
  Lextok *v;

  for (v = n->rgt; v; v = v->rgt)
    {	
      if (v->lft->ntyp == CONST)
	continue;
	
      if (v->lft->ntyp == EVAL)
	result += has_global(e, v->lft->lft);
      else
	result += has_global(e, v->lft);
    }
	
  return result;
}


static int
has_global(Element *e, Lextok *n)
{	
  Lextok *v; extern int runsafe;
  
  if (!n) return 0;
  if (AllGlobal) return 1;	/* global provided clause */

  switch (n->ntyp) {
  case '.': case BREAK: case GOTO: case CONST:
    return 0;

  case   ELSE: return n->val; /* true if combined with chan refs */

  case   NAME: return global_alias(e, n);

  case    's': return glob_args(e, n) != 0 
		 || ((n->sym->xu&(XS|XX)) != XS);

  case    'r': return glob_args(e, n) != 0 
		 || ((n->sym->xu&(XR|XX)) != XR);

  case    'R': return glob_args(e, n) != 0 
		 || (((n->sym->xu)&(XR|XS|XX)) != (XR|XS));

  case NEMPTY: return ((n->sym->xu&(XR|XX)) != XR);
  case  NFULL: return ((n->sym->xu&(XS|XX)) != XS);
  case   FULL: return ((n->sym->xu&(XR|XX)) != XR);
  case  EMPTY: return ((n->sym->xu&(XS|XX)) != XS);
  case  LEN:   return (((n->sym->xu)&(XR|XS|XX)) != (XR|XS));
    
  case RUN: return 1 - runsafe;
    
  case ENABLED: case PC_VAL: case NONPROGRESS:
  case 'p': case 'q':
  case TIMEOUT:
    return 1;

    /* 	@ was 1 (global) since 2.8.5
	in 3.0 it is considered local and
	conditionally safe, provided:
	II is the youngest process
	and nrprocs < MAXPROCS
    */
  case STOP: return 0;
  case '!': case UMIN: case '~': case ASSERT: case LVAL:
    return has_global(e, n->lft);

  case '/': case '*': case '-': case '+':
  case '%': case LT:  case GT: case '&': case '^':
  case '|': case LE:  case GE:  case NE: case '?':
  case EQ:  case OR:  case AND: case LSHIFT:
  case RSHIFT: case 'c': case ASGN:
    return has_global(e, n->lft) || has_global(e, n->rgt);

  case PRINT:
    for (v = n->lft; v; v = v->rgt)
      {
	if (has_global(e, v->lft)) 
	  return 1;
      }

    return 0;

  case NEW:
  case DELETE:
    
    return 1;
  }

  return 0;
}


static void
Bailout(FILE *fd, char *str)
{
  if (!GenCode)
    fprintf(fd, "continue%s", str);
  else 
    if (IsGuard)
      fprintf(fd, "%s%s", NextLab[Level], str);
    else
      fprintf(fd, "Uerror(\"block in step seq\")%s", str);
}


Symbol *
do_same2(Lextok *n, Symbol *v)
{
  int oln = lineno;
  Symbol *ofn = Fname;
  Lextok *fp, *tl, *tmp;

  lineno = n->ln;
  Fname  = n->fn;
  
  if (!(v->type & STRUCT))
    return v;

  if (Expand_Ok
      || v->type & POINTER)
    {
      if (!n->rgt || !n->rgt->lft)
	return v;
    }

  if (!n->rgt || !n->rgt->lft)
    fatal("no subfields '%s'", v->name);

  if (!v->Slst)
    fatal("no structure '%s'", v->name);

  if (n->rgt->ntyp != '.')
    {
      printf("bad subfield type %d\n", n->rgt->ntyp);
      alldone(1);
    }

  tmp = n->rgt->lft;
  if (tmp->ntyp != NAME && tmp->ntyp != TYPE)
    {	
      printf("bad subfield entry %d\n", tmp->ntyp);
      alldone(1);
    }

  for (fp = v->Slst; fp; fp = fp->rgt)
    for (tl = fp->lft; tl; tl = tl->rgt)
      if (!strcmp(tl->sym->name, tmp->sym->name))
	{	
	  lineno = oln; Fname = ofn;
	  return tl->sym;
	}
  
  fatal("cannot locate subfield '%s'", tmp->sym->name);

  /* never reached */
  return ZS;  
}


Symbol *
findloc2(Symbol *s)
{
  Symbol *r;
  SeqList *sl;

  extern SeqList *cur_s;

  if (!cur_s)
    {
      return ZS;
    }

  for (sl = cur_s; sl; sl = sl->nxt)
    for (r = sl->this->symtab; r; r = r->next)
      {
	if (strcmp(r->name, s->name) == 0)
	  return r;
      }

  return ZS;
}


static Symbol *
struct_last(Lextok *n, Symbol *s)
{
  Symbol *v;

  if (!s)
    return ZS;

  if (!(v = do_same2(n, s)))
    return ZS;
  
  if (!(v->type & STRUCT))
    return v;
  
  if (!(Expand_Ok || v->type & POINTER))
    {
      if (!n->rgt 
	  || !n->rgt->ntyp == '.' 
	  || !n->rgt->lft)
	fatal("unexpected struct layout `%s`", v->name);
      
      return struct_last(n->rgt->lft, v);
    }
  
  if (!n->rgt)
    return v;
  
  return struct_last(n->rgt->lft, v);
}


static Symbol *
struct_ptr(Lextok *n, Symbol *s)
{
  Symbol *v;

  if (!s)
    return ZS;

  if (!(v = do_same2(n, s)))
    return ZS;

  if (v->type & POINTER 
      || !n->rgt
      || !n->rgt->lft)
    return v;
    
  return struct_ptr(n->rgt->lft, v);
}


static Symbol *
struct_pre(Lextok *n, Symbol *s)
{  
  Symbol *v;

  if (!s)
    return ZS;

  if (!(v = do_same2(n, s)))
    return ZS;

  if (v->type != STRUCT)
    return v;

  return struct_pre(n->rgt->lft, v);
}


Lextok *
struct_tok2(Lextok *n, Symbol *s)
{
  Symbol *v;

  if (!s)
    return ZN;

    /* get the right symbol */
  if (!s->owner && (v = findloc2(s)))
    s = v;

  if (!(v = do_same2(n, s)))
    return ZN;

  if (!n->rgt)
    return n;
  
  if (!(Expand_Ok || v->type & POINTER))
    {
      if (!n->rgt->ntyp == '.' 
	  || !n->rgt->lft)
	fatal("unexpected struct layout `%s`", v->name);

      return struct_tok2(n->rgt->lft, v);
    }

  return struct_tok2(n->rgt->lft, v);
}


int
node_type2(Lextok *n)
{
  switch (n->ntyp)
    {
    case FULL:	case NFULL:
    case EMPTY:	case NEMPTY:
    case RUN:	case LEN:
      return BYTE;

    case LVAL:
      return POINTER | n->lft->sym->type;

    case __NULL:
      return POINTER;

    default:      
      if (n->ntyp == NAME)
	{
	  Symbol *s, *t;
	  Lextok *m;

	  if (!n->sym)
	    fatal("no symbol - node_type2", (char *)0);

	  if (!(s = findloc2(n->sym)))
	    s = n->sym;

	  if (!s->type)
	    return INT;

	  m = struct_tok2(n, s);
	  t = struct_last(n, s);

	  if ((t->type & POINTER) && (t->nel == 1) && m->lft)
	    return t->type & ~POINTER;

	  return t->type;
	}
      
      if (n->ntyp == FNAME)
	return FTYPE;	
       
      return INT;
    }
}


static void
putleft(FILE *fd, Lextok *now, int m)
{
  if (now->ntyp == NAME)
    putname(fd, "", now, m, "", 0);
  else
    putstmnt(fd, now, m);
}


#define cat0(x)   	putstmnt(fd,now->lft,m); \
                        fprintf(fd, x);          \
			putstmnt(fd,now->rgt,m)

#define cat1(x)		fprintf(fd,"("); cat0(x); fprintf(fd,")")
#define cat2(x,y)  	fprintf(fd,x); putstmnt(fd,y,m)
#define cat3(x,y,z)	fprintf(fd,x); putstmnt(fd,y,m); fprintf(fd,z)

static int qsend_param = 0;

void
putstmnt(FILE *fd, Lextok *now, int m)
{	
  Lextok *v;
  int i, j;

  if (!now) 
    { 
      fprintf(fd, "0"); 
      return; 
    }
	
  lineno = now->ln;
  Fname  = now->fn;

  switch (now->ntyp) 
    {
    case CONST:	fprintf(fd, "%d", now->val); break;
    case '!':	cat3(" !(", now->lft, ")"); break;
    case UMIN:	cat3(" -(", now->lft, ")"); break;
    case '~':	cat3(" ~(", now->lft, ")"); break;

    case '/':	cat1("/");  break;
    case '*':	cat1("*");  break;
    case '-':	cat1("-");  break;
    case '+':	cat1("+");  break;
    case '%':	cat1("%%"); break;
    case '&':	cat1("&");  break;
    case '^':	cat1("^");  break;
    case '|':	cat1("|");  break;
    case LT:	cat1("<");  break;
    case GT:	cat1(">");  break;
    case LE:	cat1("<="); break;
    case GE:	cat1(">="); break;

    case NE:	
      if (node_type2(now->lft) & POINTER
	  && node_type2(now->rgt) & POINTER)
	{
	  putleft(fd, now->lft, m);
	  fprintf(fd, "!=");
	  putleft(fd, now->rgt, m);
	}
      else
	{
	  cat1("!=");
	}

      break;

    case EQ:
      if (node_type2(now->lft) & POINTER
	  && node_type2(now->rgt) & POINTER)
	{
	  putleft(fd, now->lft, m);
	  fprintf(fd, "==");
	  putleft(fd, now->rgt, m);
	}
      else
	{
	  cat1("==");
	}

      break;      

    case OR:	cat1("||"); break;
    case AND:	cat1("&&"); break;
    case LSHIFT:	cat1("<<"); break;
    case RSHIFT:	cat1(">>"); break;
      
    case TIMEOUT:
      fprintf(fd, "((trpt->tau) & 1)");
      if (GenCode)
	fatal("'timeout' inside d_step sequence", (char *)0);
      break;

    case RUN:
      {
	ProcList *p;
	Lextok *lt, *fp, *ap;

	if (claimproc
	    &&  strcmp(now->sym->name, claimproc) == 0)
	  fatal("claim %s, (not runnable)", claimproc);

	if (eventmap
	    &&  strcmp(now->sym->name, eventmap) == 0)
	  fatal("eventmap %s, (not runnable)", eventmap);

	if (GenCode)
	  fatal("'run' in d_step sequence (use atomic)", (char *)0);

	p = getproc(now->sym);
	fprintf(fd,"addproc(%d\n\t\t\t", p->tn);

	Expand_Ok ++;

	for (lt = p->p, ap = now->lft, i = 0; lt && ap; lt = lt->rgt)
	  for (fp = lt->lft; fp; fp = fp->rgt, ap = ap->rgt, i ++)
	    {
	      fprintf(fd, ", ");
	      
	      if (fp->sym->type & POINTER)
		putleft(fd, ap->lft, m);
	      else
		putstmnt(fd, ap->lft, m);
	      
	      fprintf(fd, "\n\t\t\t");
	    }

	Expand_Ok --;

	if (i > Npars)
	  fatal("too many parameters in run %s(...)",
		now->sym->name);

	for ( ; i < Npars; i++)
	  fprintf(fd, ", 0");

	fprintf(fd, ")");
	break;
      }

    case NEW:
      {
	Symbol *s = now->rgt->sym;
	Symbol *l;
	SeqList *sl;

	if (!(l = findloc2(now->lft->sym)))
	  l = now->lft->sym;

	if (!l->type)
	  fatal("undeclared variable '%s'\n", l->name);

	if (!(l->type & POINTER))
	  fatal("non pointer variable '%s'\n", l->name);

	fprintf(fd, "sv_save(&now);\n");
	fprintf(fd, "#ifdef GC0\n\t\t");
	putname(fd, "gc_dec(", now->lft, m, ", 0);\n", 0);
	fprintf(fd, "#endif /* GC0 */\n");
	
	fprintf(fd, "#ifdef GC1\n\t\t");
	putname(fd, "", now->lft, m, " = 0;\n\t\t", 0);
	fprintf(fd, "gc_scan();\n\t\t");
	fprintf(fd, "gc_sweep(0);\n");
	fprintf(fd, "#endif\n\t\t");

	{ Symbol *sym;

	  lineno = now->lft->ln; 
	  Fname  = now->lft->fn;

	  if (!now->lft->sym)
	    fatal("no name - putstmnt", (char *)0);
	  
	  if (!(sym = findloc2(now->lft->sym)))
	    sym = lookup(now->lft->sym->name);

	  fprintf(fd, "{ uint x;\n");
	  fprintf(fd, "#ifdef POSYMM\n\t\t");
	  fprintf(fd, "scalar_t s;\n\t\t");
	  fprintf(fd, "symm_get_scalar(&s, II);\n\t\t");
	}

	if (s->type == STRUCT)
	  fprintf(fd, "x = addheap(%d, ", s->Nid);
	else
	  fprintf(fd, "x = addheap(%d, ", s->type);

	if (now->rgt->lft)
	  putstmnt(fd, now->rgt->lft, m);
	else
	  fprintf(fd, "1");
	
	fprintf(fd, ", &s);\n");
	fprintf(fd, "#else\n\t\t");
	if (s->type == STRUCT)
	  fprintf(fd, "x = addheap(%d, ", s->Nid);
	else
	  fprintf(fd, "x = addheap(%d, ", s->type);

	if (now->rgt->lft)
	  putstmnt(fd, now->rgt->lft, m);
	else
	  fprintf(fd, "1");
	
	fprintf(fd, ");\n");
	fprintf(fd, "#endif /* POSYMM */\n\t\t");
	putname(fd, "", now->lft, m, " = x; }\n", 0);

	fprintf(fd, "#ifdef FULLSYMM\n\t\t");
	fprintf(fd, "repnow(0);\n");
	fprintf(fd, "#endif /* FULLSYMM */\n\t\t");

	break;
      }

    case DELETE:
      { 
	SeqList *sl;

	fprintf(fd, "sv_save(&now);\n\t\t");
	putname(fd, "delheap(", now->lft, m, ");\n", 0);
	
	fprintf(fd, "#ifdef GC1\n\t\t");
	fprintf(fd, "gc_scan();\n\t\t");
	fprintf(fd, "gc_sweep(0);\n");
	fprintf(fd, "#endif\n\t\t");      
      }

      break;

    case FUNCALL:
      {
	SeqList *sl;
	Defun *df;
	int i;
	
	if (GenCode)
	  fatal("function call in d_step sequence (use atomic)", (char *)0);

	fprintf(fd, "if (!rt[II]) {\n\t\t");
	fprintf(fd, "  sv_save(&now);\n\t\t");

	if (now->rgt->ntyp != FNAME)
	  fprintf(fd, "  { uint fp;\n\n\t\t");

	Expand_Ok ++;

	for (v = now->rgt->lft, i = 0; v; v = v->rgt, i ++)
	  {
	    if (node_type2(v->lft) == STRUCT)
	      fatal("struct argument for '%s()'", now->rgt->sym->name);

	    fprintf(fd, "  Par%d = ", i);
	    putleft(fd, v->lft, m);
	    fprintf(fd, ";\n\t\t");
	  }

	fprintf(fd, "\n#ifdef GC0\n");

	for (v = now->rgt->lft; v; v = v->rgt)
	  if (v->lft && (node_type2(v->lft) & POINTER))
	    {
	      fprintf(fd, "\t\t  gc_inc(");
	      putleft(fd, v->lft, m);
	      fprintf(fd, ");\n", i);
	    }

	fprintf(fd, "#endif\n\t\t");

	Expand_Ok --;

	if (now->rgt->ntyp == FNAME)
	  {
	    if (!(df = getdefun(now->rgt->sym)))
	      fatal("defun not found (putstmnt)", 0);
	    
	    if (i > nrPars)
	      fatal("too many parameters in call %s(...)", df->fun->name);
	    
	    fprintf(fd, "  call(II, %d, start%d);\n\t\t", df->tn, df->tn);
	  }
	else
	  {
	    if (i > nrPars)
	      fatal("too many parameters in call %s(...)", 
		    now->rgt->sym->name);

	    cat3("  fp = ", now->rgt->rgt, ";\n\t\t");
	    fprintf(fd, "  cptr(II, fp);\n\t\t");
	  }

	if (now->rgt->ntyp == FNAME)
	  fprintf(fd, "  pps(II, %d", df->tn);	    
	else
	  fprintf(fd, "  pps(II, fpot(fp)");

	for (i = 0; i < nrPars; i ++)
	  fprintf(fd, ", Par%d", i);

	fprintf(fd, ");\n\t\t");
	if (now->rgt->ntyp != FNAME)
	  fprintf(fd, "  }\n\t\t");

	fprintf(fd, "  vt[II] = 1;\n\t\t");
	fprintf(fd, "  depth_adjust ++;\n\t\t");
	fprintf(fd, "}\n\t\t");

	if (now->lft)
	  {
	    fprintf(fd, "else {\n\t\t");
	    putname(fd, "  (trpt+1)->oval = ", now->lft, m, ";\n\t\t", 0);
	    putname(fd, "  ", now->lft, m, " = this->_rv;\n\t\t", 0);

	    if (node_type2(now->lft) & POINTER)
	      {
		fprintf(fd, "\n#ifdef GC0\n\t\t");
		fprintf(fd, "gc_dec((uint) (trpt+1)->oval, 1);\n");
		fprintf(fd, "#endif\n\t\t");
	      }

	    fprintf(fd, "}\n");
	  }
	
	break;
      }

    case RETFUN:
      {
	SeqList *sl;
	Defun *df;
	int i;

	if (GenCode)
	  fatal("function call in d_step sequence (use atomic)", (char *)0);

	fprintf(fd, "if (!rt[II]) {\n\t\t");
	fprintf(fd, "  sv_save(&now);\n\t\t");

	if (now->rgt->ntyp != FNAME)
	  fprintf(fd, "  uint fp;\n\n\t\t");

	Expand_Ok ++;

	for (v = now->rgt->lft, i = 0; v; v = v->rgt, i ++)
	  {
	    fprintf(fd, "  Par%d = ", i);
	    putleft(fd, v->lft, m);
	    fprintf(fd, ";\n\t\t");
	  }

	Expand_Ok --;

	fprintf(fd, "\n#ifdef GC0\n");

	for (v = now->rgt->lft; v; v = v->rgt)
	  if (v->lft && (node_type2(v->lft) & POINTER))
	    {
	      fprintf(fd, "\t\t  gc_inc(");
	      putleft(fd, v->lft, m);
	      fprintf(fd, ");\n", i);
	    }

	for (sl = now->sl; sl; sl = sl->nxt)
	  if (sl->this->sz) {
	    doformal(fd, GC0_NOSV, sl->this, 0, 0);
	    dolocal(fd, GC0_NOSV, sl->this, 0);
	  }	

	fprintf(fd, "#endif\n\t\t");

	if (now->rgt->ntyp == FNAME)
	  {
	    if (!(df = getdefun(now->rgt->sym)))
	      fatal("defun not found (putstmnt)", 0);
	    
	    if (i > nrPars)
	      fatal("too many parameters in call %s(...)", df->fun->name);
	    
	    fprintf(fd, "  poll(II, %d, start%d);\n\t\t", df->tn, df->tn);
	  }
	else
	  {
	    if (i > nrPars)
	      fatal("too many parameters in call %s(...)", 
		    now->rgt->sym->name);

	    cat3("  fp = ", now->rgt->rgt, ";\n\t\t");
	    fprintf(fd, "  polp(II, fp);\n\t\t");
	  }

	if (now->rgt->ntyp == FNAME)
	  fprintf(fd, "  pps(II, %d", df->tn);	    
	else
	  fprintf(fd, "  pps(II, fpot(fp)");

	for (i = 0; i < nrPars; i ++)
	  fprintf(fd, ", Par%d", i);
	fprintf(fd, ");\n\t\t");

	fprintf(fd, "  vt[II] = 1;\n\t\t");
	fprintf(fd, "  depth_adjust ++;\n\t\t");	
	fprintf(fd, "}\n\t\t");

	break;
      }

    case RETURN:
      {
	SeqList *l;

	fprintf(fd, "sv_save(&now);\n\t\t");
	fprintf(fd, "\n#ifdef GC0\n");

	if (now->lft
	    && now->lft->ntyp == NAME
	    && node_type2(now->lft) & POINTER)
	  putname(fd, "\t\tgc_inc(", now->lft, m, ");\n", 0);

	for (l = now->sl; l; l = l->nxt)
	  if (l->this->sz) 
	    {
	      doformal(fd, GC0_NOSV, l->this, 0, 0);
	      dolocal(fd, GC0_NOSV, l->this, 0);
	    }

	fprintf(fd, "#endif\n");
	fprintf(fd, "#ifdef GC1\n\t\t");

	if (now->lft
	    && now->lft->ntyp == NAME
	    && node_type2(now->lft) & POINTER)
	  putname(fd, "gc_mark(", now->lft, m, ");", 0);

	fprintf(fd, "\n#endif\n\t\t");

	fprintf(fd, "ret(II, ");
	if (RetType && (RetType->type & POINTER))
	  {
	    putleft(fd, now->lft, m);
	    fprintf(fd, ", 1);\n\t\t");
	  }
	else 
	  {
	    putstmnt(fd, now->lft, m);
	    if (RetType && (RetType->type == CHAN))
	      fprintf(fd, ", 2);\n\t\t");
	    else
	      fprintf(fd, ", 0);\n\t\t");
	  }

	fprintf(fd, "rt[II] = 1;\n\t\t");
	fprintf(fd, "vt[II] = 1;\n\t\t");

	fprintf(fd, "\n#ifdef GC1\n\t\t");
	fprintf(fd, "gc_scan();\n");
	fprintf(fd, "#ifdef FULLSYMM\n\t\t");
	fprintf(fd, "{ int p[MAXHEAP];\n\t\t");
	fprintf(fd, "  int l = copy_perm(p, 0);\n\t\t");
	fprintf(fd, "  perm_heap(p, l, 0); }\n");
	fprintf(fd, "#endif /* FULLSYMM */\n\t\t");
	fprintf(fd, "gc_sweep(0);\n");
	fprintf(fd, "#endif /* GC1 */\n\t\t");
      }

      break;

    case ENABLED:
      cat3("enabled(II, ", now->lft, ")");
      break;

    case NONPROGRESS:
      /* o_pm&4=progress, tau&128=claim stutter */
      fprintf(fd, "(!(trpt->o_pm&4) && !(trpt->tau&128))");
      break;

    case PC_VAL:
      cat3("pptr(", now->lft, "+ BASE)->_p");
      break;

    case LEN:
      if (!terse && !TestOnly && has_xu)
	{	
	  fprintf(fd, "\n#ifndef XUSAFE\n\t\t");
	  putname(fd, "(!(q_claim[", now->lft, m, "]&1) || ", 1);
	  putname(fd, "q_R_check(", now->lft, m, "", 1);
	  fprintf(fd, ", II)) &&\n\t\t");
	  putname(fd, "(!(q_claim[", now->lft, m, "]&2) || ", 1);
	  putname(fd, "q_S_check(", now->lft, m, ", II)) &&", 1);
	  fprintf(fd, "\n#endif\n\t\t");
	}
	
      putname(fd, "q_len(", now->lft, m, ")", 1);
      break;

    case FULL:
      if (!terse && !TestOnly && has_xu)
	{	
	  fprintf(fd, "\n#ifndef XUSAFE\n\t\t");
	  putname(fd, "(!(q_claim[", now->lft, m, "]&1) || ", 1);
	  putname(fd, "q_R_check(", now->lft, m, "", 1);
	  fprintf(fd, ", II)) &&\n\t\t");
	  putname(fd, "(!(q_claim[", now->lft, m, "]&2) || ", 1);
	  putname(fd, "q_S_check(", now->lft, m, ", II)) &&", 1);
	  fprintf(fd, "\n#endif\n\t\t");
	}
      
      putname(fd, "q_full(", now->lft, m, ")", 1);
      break;

    case EMPTY:
      if (!terse && !TestOnly && has_xu)
	{	
	  fprintf(fd, "\n#ifndef XUSAFE\n\t\t");
	  putname(fd, "(!(q_claim[", now->lft, m, "]&1) || ", 1);
	  putname(fd, "q_R_check(", now->lft, m, "", 1);
	  fprintf(fd, ", II)) &&\n\t\t");
	  putname(fd, "(!(q_claim[", now->lft, m, "]&2) || ", 1);
	  putname(fd, "q_S_check(", now->lft, m, ", II)) &&", 1);
	  fprintf(fd, "\n#endif\n\t\t");
	}
      
      putname(fd, "(q_len(", now->lft, m, ")==0)", 1);
      break;

    case NFULL:
      if (!terse && !TestOnly && has_xu)
	{	
	  fprintf(fd, "\n#ifndef XUSAFE\n\t\t");
	  putname(fd, "(!(q_claim[", now->lft, m, "]&2) || ", 1);
	  putname(fd, "q_S_check(", now->lft, m, ", II)) &&", 1);
	  fprintf(fd, "\n#endif\n\t\t");
	}
      
      putname(fd, "(!q_full(", now->lft, m, "))", 1);
      break;

    case NEMPTY:
      if (!terse && !TestOnly && has_xu)
	{	
	  fprintf(fd, "\n#ifndef XUSAFE\n\t\t");
	  putname(fd, "(!(q_claim[", now->lft, m, "]&1) || ", 1);
	  putname(fd, "q_R_check(", now->lft, m, ", II)) &&", 1);
	  fprintf(fd, "\n#endif\n\t\t");
	}
	
      putname(fd, "(q_len(", now->lft, m, ")>0)", 1);
      break;

    case 's':
      if (Pid == eventmapnr)
	{	
	  fprintf(fd, "if (_tp != 's' ");
	  putname(fd, "|| _qid + 1 != ", now->lft, m, "", 1);

	  for (v = now->rgt, i=0; v; v = v->rgt, i++)
	    {	
	      if (v->lft->ntyp != CONST
		  &&  v->lft->ntyp != EVAL)
		continue;
		
	      fprintf(fd, " \\\n\t\t|| qrecv(");
	      putname(fd, "", now->lft, m, ", ", 1);
	      putname(fd, "q_len(", now->lft, m, ")-1, ", 1);
	      fprintf(fd, "%d, 0) != ", i);
	      if (v->lft->ntyp == CONST)
		putstmnt(fd, v->lft, m);
	      else /* EVAL */
		putstmnt(fd, v->lft->lft, m);
	    }
		
	  fprintf(fd, ")\n");
	  fprintf(fd, "\t\t	continue");
	  putname(th, " || (x_y3_ == ", now->lft, m, ")", 1);
	  break;
	}
	
      if (TestOnly)
	{	
	  if (m_loss)
	    fprintf(fd, "1");
	  else
	    putname(fd, "!q_full(", now->lft, m, ")", 1);
	  break;
	}
	
      if (has_xu)
	{	
	  fprintf(fd, "\n#ifndef XUSAFE\n\t\t");
	  putname(fd, "if (q_claim[", now->lft, m, "]&2) ", 1);
	  putname(fd, "q_S_check(", now->lft, m, ", II);", 1);
	  fprintf(fd, "\n#endif\n\t\t");
	}
	
      fprintf(fd, "if (q_%s", (u_sync > 0 && u_async == 0)?"len":"full");
      putname(fd, "(", now->lft, m, "))\n", 1);

      if (m_loss)
	{	
	  if (!GenCode)
	    fprintf(fd, "\t\t{ nlost++; m=3; goto P999; }\n");
	  else
	    fprintf(fd, "\t\t\tnlost++;\n\t\telse {");
	} 
      else {	
	fprintf(fd, "\t\t\t");
	Bailout(fd, ";");
      }

      if (has_enabled)
	fprintf(fd, "\n\t\tif (TstOnly) return 1;");
      
      putname(fd, "\n\t\tqsend(", now->lft, m, "", 1);
      fprintf(fd, ", %d", now->val);

      qsend_param = 1;
      for (v = now->rgt, i = 0; v; v = v->rgt, i++)
	{	
	  cat2(", ", v->lft);
	}
      qsend_param = 0;

      if (i > Mpars)
	{	
	  putname(stdout, "channel name: ", now->lft, m, "\n", 1);
	  printf("	%d msg parameters\n", i);
	  fatal("too many pars in send", "");
	}
	
      for ( ; i < Mpars; i++)
	fprintf(fd, ", 0");
      
      fprintf(fd, ")");

      if (u_sync) {	
	fprintf(fd, ";\n\t\t");
	if (u_async)
	  putname(fd, "if (q_zero(", now->lft, m, ")) ", 1);
	
	putname(fd, "{ boq = ", now->lft, m, "", 1);
	
	if (GenCode)
	  fprintf(fd, "; Uerror(\"rv-attempt in d_step\")");
	
	fprintf(fd, "; }");
      }
	
      if (m_loss && GenCode)
	fprintf(fd, ";\n\t\t}\n\t\t");
      break;

    case 'r':
      if (Pid == eventmapnr)
	{	
	  fprintf(fd, "if (_tp != 'r' ");
	  putname(fd, "|| _qid+1 != ", now->lft, m, "", 1);
	  
	  for (v = now->rgt, i=0; v; v = v->rgt, i++)
	    {	
	      if (v->lft->ntyp != CONST
		  &&  v->lft->ntyp != EVAL)
		continue;
		
	      fprintf(fd, " \\\n\t\t|| qrecv(");
	      putname(fd, "", now->lft, m, ", ", 1);
	      fprintf(fd, "0, %d, 0) != ", i);
	      if (v->lft->ntyp == CONST)
		putstmnt(fd, v->lft, m);
	      else /* EVAL */
		putstmnt(fd, v->lft->lft, m);
	    }
		
	  fprintf(fd, ")\n");
	  fprintf(fd, "\t\t	continue");

	  putname(tc, " || (x_y3_ == ", now->lft, m, ")", 1);
	  
	  break;
	}
	
      if (TestOnly)
	{	
	  fprintf(fd, "((");
	  
	  if (u_sync)
	    fprintf(fd, "(boq == -1 && ");
		
	  putname(fd, "q_len(", now->lft, m, ")", 1);
	  if (u_sync && now->val <= 1)
	    { 
	      putname(fd, ") || (boq == ",  now->lft,m," && ", 1);
	      putname(fd, "q_zero(", now->lft,m,"))", 1);
	    }
		
	  fprintf(fd, ")");
	  if (now->val == 0 || now->val == 2)
	    {	
	      for (v = now->rgt, i=j=0; v; v = v->rgt, i++)
		{ 
		  if (v->lft->ntyp == CONST)
		    { 
		      cat3("\n\t\t&& (", v->lft, " == ");
		      putname(fd, "qrecv(", now->lft, m, ", ", 1);
		      fprintf(fd, "0, %d, 0))", i);
		    } else 
		      if (v->lft->ntyp == EVAL)
			{ 
			  cat3("\n\t\t&& (", v->lft->lft, " == ");
			  putname(fd, "qrecv(", now->lft, m, ", ", 1);
			  fprintf(fd, "0, %d, 0))", i);
			} 
		      else {	
			j++; 
			continue;
		      }
		}
			
	      fprintf(fd, ")");
	    } 
	  else {	
	    fprintf(fd, "\n\t\t&& Q_has(");
	    putname(fd, "", now->lft, m, "", 1);
	    
	    for (v = now->rgt, i=0; v; v = v->rgt, i++)
	      {	
		if (v->lft->ntyp == CONST)
		  {	
		    fprintf(fd, ", 1, ");
		    putstmnt(fd, v->lft, m);
		  } else 
		    if (v->lft->ntyp == EVAL)
		      {	
			fprintf(fd, ", 1, ");
			putstmnt(fd, v->lft->lft, m);
		      } 
		    else {	
		      fprintf(fd, ", 0, 0");
		    }	
	      }
		
	    for ( ; i < Mpars; i++)
	      fprintf(fd, ", 0, 0");
	    
	    fprintf(fd, "))");
	  }

	  break;
	}
	
      if (has_xu)
	{	
	  fprintf(fd, "\n#ifndef XUSAFE\n\t\t");
	  putname(fd, "if (q_claim[", now->lft, m, "]&1) ", 1);
	  putname(fd, "q_R_check(", now->lft, m, ", II);", 1);
	  fprintf(fd, "\n#endif\n\t\t");
	}
      
      if (u_sync)
	{	
	  if (now->val >= 2)
	    {	
	      if (u_async)
		{ 
		  fprintf(fd, "if (");
		  putname(fd, "q_zero(", now->lft,m,"))", 1);
		  fprintf(fd, "\n\t\t{\t");
		}
		
	      fprintf(fd, "uerror(\"polling ");
	      fprintf(fd, "rv chan\");\n\t\t");
	      
	      if (u_async)
		fprintf(fd, "	continue;\n\t\t}\n\t\t");

	      fprintf(fd, "IfNotBlocked\n\t\t");
	    } 
	  else {	
	    fprintf(fd, "if (");
	    
	    if (u_async == 0)
	      putname(fd, "boq != ", now->lft,m,") ", 1);
	    else { 
	      putname(fd, "q_zero(", now->lft,m,"))", 1);
	      fprintf(fd, "\n\t\t{\tif (boq != ");
	      putname(fd, "",  now->lft,m,") ", 1);
	      Bailout(fd, ";\n\t\t} else\n\t\t");
	      fprintf(fd, "{\tif (boq != -1) ");
	    }
		
	    Bailout(fd, ";\n\t\t");
	    if (u_async)
	      fprintf(fd, "}\n\t\t");
	  }	
	}
	
      putname(fd, "if (q_len(", now->lft, m, ") == 0) ", 1);
      Bailout(fd, "");
      
      for (v = now->rgt, j=0; v; v = v->rgt)
	{	
	  if (v->lft->ntyp != CONST
	      &&  v->lft->ntyp != EVAL)
	    j++;	/* count settables */
	}
	
      fprintf(fd, ";\n\t\t{\n\t\tint XX=1");
      /* test */      
      if (now->val == 0 || now->val == 2)
	{	
	  for (v = now->rgt, i=0; v; v = v->rgt, i++)
	    {	
	      if (v->lft->ntyp == CONST)
		{ 
		  fprintf(fd, ";\n\t\t");
		  cat3("if (", v->lft, " != ");
		  putname(fd, "qrecv(", now->lft, m, ", ", 1);
		  fprintf(fd, "0, %d, 0)) ", i);
		  Bailout(fd, "");
		} else 
		  if (v->lft->ntyp == EVAL)
		    { 
		      fprintf(fd, ";\n\t\t");
		      cat3("if (", v->lft->lft, " != ");
		      putname(fd, "qrecv(", now->lft, m, ", ", 1);
		      fprintf(fd, "0, %d, 0)) ", i);
		      Bailout(fd, "");
		    }	
	    }
	} 
      else {	/* random receives */
	fprintf(fd, ";\n\t\tif (!(XX = Q_has(");
	putname(fd, "", now->lft, m, "", 1);
	
	for (v = now->rgt, i=0; v; v = v->rgt, i++)
	  {	
	    if (v->lft->ntyp == CONST)
	      {	
		fprintf(fd, ", 1, ");
		putstmnt(fd, v->lft, m);
	      } else 
		if (v->lft->ntyp == EVAL)
		  {	
		    fprintf(fd, ", 1, ");
		    putstmnt(fd, v->lft->lft, m);
		  } 
		else
		  fprintf(fd, ", 0, 0");
	  }
		
	for ( ; i < Mpars; i++)
	  fprintf(fd, ", 0, 0");
	
	fprintf(fd, "))) ");
	Bailout(fd, "");
      }

      if (has_enabled)
	fprintf(fd, ";\n\t\tif (TstOnly) return 1");

      if (j == 0 && now->val >= 2)
	{	
	  fprintf(fd, ";\n\t}\n\t\t");
	  break;	/* poll without side-effect */
	}

      if (!GenCode)
	{	
	  fprintf(fd, ";\n");

	  if (j == 0 && now->val == 0)
	    {	
	      fprintf(fd, "\t\tif (q_flds[qptr(");
	      putname(fd, "", now->lft, m, "-1)->_t]", 1);
	      fprintf(fd, " != %d)\n", i);
	    }
	
	  /*
	   * fprintf(fd, "#if VECTORSZ<=1024\n");
	   * fprintf(fd, "#if defined(FULLSTACK)");
	   * fprintf(fd, " && defined(NOCOMP)");
	   * fprintf(fd, " && !defined(BITSTATE)");
	   * fprintf(fd, " && !defined(MA)\n");
	   * fprintf(fd, "\t\tif (t->atom & 2)\n");
	   * fprintf(fd, "#endif\n");
	   * fprintf(fd, "\t\t\tsv_save(&now);\n");
	   * fprintf(fd, "#else\n");
	   */

	  fprintf(fd, "\t\t\tp_q_save(II, ");
	  putname(fd, "", now->lft, m, ");\n", 1);

	  /*
	   * fprintf(fd, "#endif\n\t\t");	  
	   */
	}

      /* set */	
      for (v = now->rgt, i = 0; v; v = v->rgt, i++)
	{	
	  if ((v->lft->ntyp == CONST
	       ||   v->lft->ntyp == EVAL) && v->rgt)
	    continue;
		
	  fprintf(fd, ";\n\t\t");
	  if (v->lft->ntyp != CONST
	      &&  v->lft->ntyp != EVAL)
	    {	
	      Symbol *lsym = v->lft->sym;

	      if (!lsym)
		fatal("no left value - putstmnt (r)", (char *)0);

	      if (lsym->context && context && lsym->type)
		lsym = lookup(lsym->name);

	      if (!lsym->type)
		lsym = lookup(lsym->name);

	      if (!lsym->type)
		fatal("undeclared variable '%s'\n", lsym->name);

	      _isok ++;
	      putname(fd, "", v->lft, m, "", 0);
	      _isok --;
	      
	      if (node_type2(v->lft) == BIT)
		{
		  putname(fd, "qrecv(", now->lft, m, ", ", 1);
		  fprintf(fd, "XX-1, %d, ", i);
		  fprintf(fd, "%d))", (v->rgt || now->val >= 2) ? 0 : 1);
		}
	      else {
		fprintf(fd, " = ");
		putname(fd, "qrecv(", now->lft, m, ", ", 1);
		fprintf(fd, "XX-1, %d, ", i);
		fprintf(fd, "%d)", (v->rgt || now->val >= 2) ? 0 : 1); 
	      }
	    }
	  else {
	    putname(fd, "qrecv(", now->lft, m, ", ", 1);
	    fprintf(fd, "XX-1, %d, ", i);
	    fprintf(fd, "%d)", (v->rgt || now->val >= 2) ? 0 : 1);	  
	  }
	}
	
      fprintf(fd, ";\n\t}\n\t\t");
      if (u_sync)
	{	
	  putname(fd, "if (q_zero(", now->lft, m, "))", 1);
	  fprintf(fd, "\n\t\t{	boq = -1;\n");
	  
	  fprintf(fd, "#ifndef NOFAIR\n"); /* NEW 3.0.8 */
	  fprintf(fd, "\t\t\tif (fairness\n");
	  fprintf(fd, "\t\t\t&& !(trpt->o_pm&32)\n");
	  fprintf(fd, "\t\t\t&& (now._a_t&2)\n");
	  fprintf(fd, "\t\t\t&&  now._cnt[now._a_t&1] == II+2)\n");
	  fprintf(fd, "\t\t\t{	now._cnt[now._a_t&1] -= 1;\n");
	  fprintf(fd, "#ifdef VERI\n");
	  fprintf(fd, "\t\t\t	if (II == 1)\n");
	  fprintf(fd, "\t\t\t		now._cnt[now._a_t&1] = 1;\n");
	  fprintf(fd, "#endif /* VERI */\n");
	  fprintf(fd, "\t\t\t	trpt->o_pm |= (32|64);\n");
	  fprintf(fd, "\t\t\t}\n");
	  fprintf(fd, "#endif /* NOFAIR */\n");
	  
	  fprintf(fd, "\n\t\t}");
	}
      break;
      
    case 'R':
      if (!terse && !TestOnly && has_xu)
	{	
	  fprintf(fd, "\n#ifndef XUSAFE\n\t\t");
	  putname(fd, "(!(q_claim[", now->lft, m, "]&1) || ", 1);
	  fprintf(fd, "q_R_check(");
	  putname(fd, "", now->lft, m, ", II)) &&\n\t\t", 1);
	  putname(fd, "(!(q_claim[", now->lft, m, "]&2) || ", 1);
	  putname(fd, "q_S_check(", now->lft, m, ", II)) &&", 1);
	  fprintf(fd, "\n#endif\n\t\t");
	}
	
      if (u_sync>0)
	putname(fd, "not_RV(", now->lft, m, ") && \\\n\t\t", 1);
      
      for (v = now->rgt, i=j=0; v; v = v->rgt, i++)
	if (v->lft->ntyp != CONST
	    &&  v->lft->ntyp != EVAL)
	  {	
	    j++; 
	    continue;
	  }
      
      if (now->val == 0 || i == j)
	{	
	  putname(fd, "(q_len(", now->lft, m, ") > 0", 1);

	  for (v = now->rgt, i=0; v; v = v->rgt, i++)
	    {	
	      if (v->lft->ntyp != CONST
		  &&  v->lft->ntyp != EVAL)
		continue;
	      
	      fprintf(fd, " \\\n\t\t&& qrecv(");
	      putname(fd, "", now->lft, m, ", ", 1);
	      fprintf(fd, "0, %d, 0) == ", i);
	      
	      if (v->lft->ntyp == CONST)
		putstmnt(fd, v->lft, m);
	      else /* EVAL */
		putstmnt(fd, v->lft->lft, m);
	    }
	  fprintf(fd, ")");
	} 
      else {	
	putname(fd, "Q_has(", now->lft, m, "", 1);

	for (v = now->rgt, i=0; v; v = v->rgt, i++)
	  {	
	    if (v->lft->ntyp == CONST)
	      {	
		fprintf(fd, ", 1, ");
		putstmnt(fd, v->lft, m);
	      } else 
		if (v->lft->ntyp == EVAL)
		  {	
		    fprintf(fd, ", 1, ");
		    putstmnt(fd, v->lft->lft, m);
		  } 
		else
		  fprintf(fd, ", 0, 0");
	  }	
		
	for ( ; i < Mpars; i++)
	  fprintf(fd, ", 0, 0");
	
	fprintf(fd, ")");
      }
      
      break;

    case 'c':
      cat3("if (!(", now->lft, "))\n\t\t\t");
      Bailout(fd, "");
      break;
      
    case  ELSE:
      if (!GenCode)
	{	
	  fprintf(fd, "if (trpt->o_pm&1)\n\t\t\t");
	  Bailout(fd, "");
	} 
      else {	
	fprintf(fd, "/* else */");
      }
      
      break;

    case '?':
      cat3("( (", now->lft, ") ? ");
      cat3("(", now->rgt->lft, ") : ");
      cat3("(", now->rgt->rgt, ") )");

      break;

    case ASGN:
      {
	Symbol *lsym;
	int lt, rt;
	char pre[128], suff[128];

	if (has_enabled)
	  fprintf(fd, "if (TstOnly) return 1;\n\t\t");
      
	if (!now->lft->sym)
	  fatal("no left value - putstmnt (ASGN)", (char *)0);

	if (!(lsym = findloc2(now->lft->sym)))
	  lsym = now->lft->sym;

	if (!lsym->type)
	  fatal("undeclared variable '%s'\n", lsym->name);

	lt = node_type2(now->lft);
	rt = node_type2(now->rgt);

	if (!typ_ck(lt, rt, "assignment"))
	  alldone(1);

	_isok ++;

	if ((lt & POINTER) && (rt & POINTER))
	  {
	    fprintf(fd, "\n#ifdef GC0\n\t\t");
	    putname(fd, "gc_dec(", now->lft, m, ", 1);\n", 0);
	    fprintf(fd, "#endif\n\t\t");
	  }

	if (!GenCode)
	  {
	    fprintf(fd, "(trpt+1)->oval =\n\t\t");
	    
	    if ((lt & POINTER) && (rt & POINTER))
	      putname(fd, "", now->lft, m, "", 0);
	    else
	      putname(fd, "", now->lft, m, "", 1);

	    fprintf(fd, ";\n\t\t");
	  }

	if (lt == BIT)
	  {
	    putname (fd, "", now->lft, m, "", 0);
	  }
	else
	  {
	    if ((lt & POINTER) && (rt & POINTER))
	      putname(fd, "", now->lft, m, "", 0);
	    else
	      putname(fd, "", now->lft, m, "", 1);
	  }

	if (lt == BIT)
	  {
	    putstmnt(fd, now->rgt, m);
	    fprintf(fd, ")");
	  }
	else
	  {
	    fprintf(fd," = ");

	    if ((now->rgt->ntyp == NAME) &&
		(lt & POINTER) && (rt & POINTER))
	      putname(fd, "", now->rgt, m, "", 0);
	    else
	      putstmnt(fd, now->rgt, m);
	  }

	fprintf(fd, ";\n\t\t");

	if ((lt & POINTER) && (rt & POINTER))
	  {
	    if (now->rgt->ntyp == NAME)
	      {
		fprintf(fd, "#ifdef GC0\n\t\t");
		putname(fd, "gc_inc(", now->lft, m, ");\n", 0);
		fprintf(fd, "#endif /* GC0 */\n");
	      }

	    fprintf(fd, "#ifdef GC1\n\t\t");
	    fprintf(fd, "gc_scan();\n");
	    fprintf(fd, "#ifdef FULLSYMM\n\t\t");
	    fprintf(fd, "{ int p[MAXHEAP];\n\t\t");
            fprintf(fd, "  int l = copy_perm(p, 0);\n\t\t");
	    fprintf(fd, "  perm_heap(p, l, 1); }\n");
	    fprintf(fd, "#endif /* FULLSYMM */\n\t\t");
	    fprintf(fd, "  gc_sweep(1);\n");
	    fprintf(fd, "#endif /* GC1 */\n\t\t");
	  }

	_isok--;	
      }
	
      break;

    case PRINT:
      if (has_enabled)
	fprintf(fd, "if (TstOnly) return 1;\n\t\t");
    
      fprintf(fd, "\n#ifdef PRINTF\n\t\t");
      fprintf(fd, "printf(%s", now->sym->name);

      for (v = now->lft; v; v = v->rgt)
	{	
	  cat2(",\n\t\t\t", v->lft);
	}
      
      fprintf(fd, ");\n");
      fprintf(fd, "#endif\n\t\t");

      break;

    case FNAME:
      {
	Defun *df;

	if (!(df = getdefun(now->sym)))
	  fatal("defun not found (putstmnt)", 0);

	fprintf(fd, "fptr(%d, start%d)", df->tn, df->tn);
      }

      break;

    case NAME:
      putname(fd, "", now, m, "", !qsend_param);
      break;

    case LVAL:
      {
	Symbol *s;

	if (!(s = findloc2(now->lft->sym)))
	  s = now->lft->sym;

	if (!s->type)
	  fatal("undeclared variable '%s'\n", s->name);

	fprintf(fd, "mp(");
	putoffset(fd, now->lft, s, m, 0, 0);
	fprintf(fd, ", ");

	if (s->context)
	  fprintf(fd, "II, LPTR)");
	else
	  fprintf(fd, "0, GPTR)");

	break;
      }

    case __NULL:
      fprintf(fd, "null");
      break;

    case   'p':
      putremote(fd, now, m);
      break;

    case   'q':
      if (terse)
	fprintf(fd, "%s", now->sym->name);
      else
	fprintf(fd, "%d", remotelab(now));
      break;

    case ASSERT:
      {
	extern void comwork(FILE *, Lextok *, int);

	if (has_enabled)
	  fprintf(fd, "if (TstOnly) return 1;\n\t\t");

	cat3("assert(", now->lft, ", ");
	terse = nocast = 1;

	/* need to print the symbolic statement */
	fprintf(fd, "\"");
	comwork(fd, now->lft, m);
	fprintf(fd, "\", II, tt, t)");
	terse = nocast = 0;

	break;
      }

    case '.':
    case BREAK:
    case GOTO:
      if (Pid == eventmapnr)
	fprintf(fd, "Uerror(\"cannot get here\")");
	
      putskip(m);
      break;

    case STOP: /* Stop */     

      if (Pid == eventmapnr)
	{ 
	  fprintf(fd, "return 0");
	  break;
	}

      if (has_enabled)
	{	
	  fprintf(fd, "if (TstOnly)\n\t\t");
	  fprintf(fd, "  return (II+1 == now._nr_pr);\n\t\t");
	}

      fprintf(fd, "if (now._nr_pr == II + 1) {\n\t\t");

      fprintf(fd, "\n#ifdef GC0\n");
      doformal(fd, GC0_SAVE, Proc->s, 0, 0);
      dolocal(fd, GC0_SAVE, Proc->s, 0);
      fprintf(fd, "#endif\n\t\t");

      fprintf(fd, "  delproc(1, II);\n");

      fprintf(tm, "#ifdef GC1\n\t\t");
      fprintf(tm, "gc_scan();\n\t\t");
      fprintf(tm, "gc_sweep(1);\n");
      fprintf(tm, "#endif\n\t\t");

      fprintf(fd, "}\n\t\t");
      fprintf(fd, "else ");
      Bailout(fd, "");

      break;

    default:
      printf("spin: bad node type %d `%c' (.m)\n", now->ntyp, now->ntyp);
      alldone(1);
    }
}


/*
 * Recursive back-end for putname(). Computes the 
 * actual offset and C type for a variable reference.
 */

static void
putoffset(FILE *fd, Lextok *n, Symbol *s, int m, long base, int puttype) 
{
  long newb = 0;

  if (s->nel > 1) 
    {	
      /* an xr or xs reference to an array element
       * becomes an exclusion tag on the array itself -
       * which could result in invalidly labeling
       * operations on other elements of this array to
       * be also safe under the partial order reduction
       * (see procedure has_global())
       */

      if (no_arrays)
	{
	  non_fatal("ref to array element invalid in this context", (char *)0);
	  printf("\thint: instead of, e.g., x[rs] qu[3], use\n");
	  printf("\tchan nm_3 = qu[3]; x[rs] nm_3;\n");
	  printf("\tand use nm_3 in sends/recvs instead of qu[3]\n");
	}

      if (n->lft
	  && n->lft->ntyp == CONST
	  && n->lft->val < s->nel)
	newb = base + s->offset + (s->bitpos + n->lft->val * s->bitsize) / 8;
      else 
	{
	  if (!n->lft && s->nel > 0)
	    newb = base + s->offset;
	  else { 
	    fprintf(fd, "%d+(%d+%d*", base + s->offset, s->bitpos, s->bitsize);
	    cat3("Index(", n->lft, ", ");
	    fprintf(fd, "%d))/8+", s->nel);
	  }
	}
    }
  else
    newb = base + s->offset;

  if (s->type == STRUCT
      && n->rgt && n->rgt->lft)
    {
      putoffset(fd, n->rgt->lft, 
		do_same2(n, n->sym),
		m, newb, puttype);
      return;
    }

  if (puttype)
    fprintf(fd, "%d, %s", newb, typ2c(s));
  else
    fprintf(fd, "%d", newb);
}


/*
 * Computes the bitpos for bit variables.
 */

static void
putbitpos(FILE *fd, Symbol *s, Lextok *n, int m)
{
  if (n
      && n->ntyp == CONST
      && n->val < s->nel)
    fprintf(fd, "%d", (s->bitpos + n->val) % 8);
  else
    {
      if (!n && s->nel > 0)
	fprintf(fd, "%d", s->bitpos);
      else {
	fprintf(fd, "(%d+", s->bitpos);
	cat3("Index(", n, ", ");
	fprintf(fd, "%d))%%8", s->nel);
      }
    }
}


/*
 * putname() front-end. Prints all dp's and xx's at the 
 * beginning of the expression in the reversed order of the fields.
 */

static void
__putname1(FILE *fd, Lextok *n, Symbol *s, int rval)
{
  if ((s->type & STRUCT) && n->rgt && n->rgt->lft)
    {
      __putname1(fd, n->rgt->lft, do_same2(n, n->sym), rval);      
      if (s->type & POINTER) 
	{
	  fprintf(fd, "*(%s *)xx(dp(", dep2c(struct_pre(n, s)));
	}
    }
  else {
    if ((s->type & POINTER) && rval)
      {
	if (s->nel == 1 && n->lft)
	  fprintf(fd, "*(%s *)xx(dp(", dep2c(struct_pre(n, s)));
	else
	  fprintf(fd, "*(%s *)(dp(", dep2c(s));
      }

    if (s->type == BIT)
      {
	if (rval)
	  fprintf(fd, "bget(");
	else
	  fprintf(fd, "bset(");
      }

    return;
  }
}


/*
 * putname() back-end. Closes all paranthesis opened by __putname1().
 * Uses the normal order in which fields appear.
 */
static void
__putname2(FILE *fd, Lextok *n, Symbol *s, int m, int rval)
{
  if ((s->type & POINTER) && rval)
    {
      fprintf(fd, ", ");
      if (s->nel == 1 && n->lft) 
	{
	  putstmnt(fd, n->lft, m);
	  fprintf(fd, "*sizeof(%s)", dep2c(s));
	}
      else
	fprintf(fd, "0");
      
      if ((s->type & STRUCT) && n->rgt && n->rgt->lft)
	fprintf(fd, "+");
      else
	fprintf(fd, "))");
    }

  if (s->type == BIT)
    {
      fprintf(fd, ",\n\t\t\t");
      putbitpos(fd, s, n->lft, m);

      if (rval)
	fprintf(fd, ")");
      else
	fprintf(fd, ", ");
    }

  if ((s->type & STRUCT) && n->rgt && n->rgt->lft)
    {
      if (s->type & POINTER)
	{
	  if (!rval)
	    fprintf(fd, ", ");

	  putoffset(fd, n->rgt->lft, do_same2(n, n->sym), m, 0, 0);
	  fprintf(fd, "))");
	}

      __putname2(fd, n->rgt->lft, do_same2(n, n->sym), m, rval);
    }
}


void
putname(FILE *fd, char *pre, Lextok *n, int m, char *suff, int rval)
{
  Symbol *s;
  Symbol *t; /* structure last symbol */
  
  lineno = n->ln; 
  Fname  = n->fn;

  if (!n->sym)
    fatal("no name - putname", (char *)0);

  if (!(s = findloc2(n->sym)))
    s = lookup(n->sym->name);

  if (!s)
    {	
      fprintf(fd, "%s%s%s", pre, n->sym->name, suff);
      return;  
    }

  if (!s->type)
    fatal("undeclared variable '%s'", s->name);

  if (s->type == PROCTYPE)
    fatal("proctype-name '%s' used as array-name", s->name);

  fprintf(fd, pre);
  
  if (strcmp(s->name, "_p") == 0 ||
      strcmp(s->name, "_pid") == 0)
    {
      fprintf(fd, "this->%s%s", s->name, suff);
      return;
    }

  if (strcmp(s->name, "_") == 0)
    {
      if (_isok == 0)
	fatal("attempt to read value of '_'", 0);
      else
	fprintf(fd, "_%s", suff);
      
      return;
    }

  __putname1(fd, n, s, rval);

  if (s->hidden & 1)
    {
      fprintf(fd, "%s", s->name);

      if (s->nel > 1)
	{
	  fprintf(fd, "[");

	  if (n->lft)
	    putstmnt(fd, n->lft, m);
	  else
	    fprintf(fd, "0");

	  fprintf(fd, "]");
	}

      fprintf(fd, "%s", suff);
      return;
    }

  t = struct_ptr(n, s);

  if (s->context)
    {
      if (t->type == STRUCT)
	fprintf(fd, "lptr(");
      else
	fprintf(fd, "ll(");
    }
  else
    {
      if (t->type == STRUCT)
	fprintf(fd, "gptr(");
      else
	fprintf(fd, "gg(");
    }

  putoffset(fd, n, s, m, 0, t->type != STRUCT);
  fprintf(fd, ")");

  __putname2(fd, n, s, m, rval);

  fprintf(fd, "%s", suff);
}


void
putremote(FILE *fd, Lextok *n, int m)	/* remote reference */
{	
  int promoted = 0;

  if (terse)
    {	
      fprintf(fd, "%s[", n->lft->sym->name);
      putstmnt(fd, n->lft->lft, m);
      fprintf(fd, "].%s", n->sym->name);
    } 
  else {
    if (node_type(n) < SHORT)
      {	
	promoted = 1;
	fprintf(fd, "((int)");
      }
    
    fprintf(fd, "pptr(", fproc(n->lft->sym->name));
    fprintf(fd, "BASE + ");
    putstmnt(fd, n->lft->lft, m);
    fprintf(fd, ")->%s", n->sym->name);
  }
  
  if (n->rgt)
    {	
      fprintf(fd, "[");
      putstmnt(fd, n->rgt, m);
      fprintf(fd, "]");
    }
  
  if (promoted) 
    fprintf(fd, ")");
}



static int
getweight(Lextok *n)
{
	switch (n->ntyp) {
	case 'r':     return 4;
	case 's':     return 2;
	case TIMEOUT: return 1;	/* lowest priority */
	case 'c':     if (has_typ(n->lft, TIMEOUT)) return 1;
	}
	return 3;
}


int
has_typ(Lextok *n, int m)
{
	if (!n) return 0;
	if (n->ntyp == m) return 1;
	return (has_typ(n->lft, m) || has_typ(n->rgt, m));
}

